import numpy as np
import time
from scipy.optimize import minimize_scalar
import logging


from xframe.library.pythonLibrary import DictNamespace
from xframe.library.pythonLibrary import create_threshold_projection
from xframe.library.physicsLibrary import spherical_formfactor
from xframe.library.gridLibrary import NestedArray,GridFactory
from xframe.library.gridLibrary import ReGrider,SampledFunction
from xframe.library.mathLibrary import PolarIntegrator,SphericalIntegrator,distance_from_line_2d,midpoint_rule
from xframe.library.mathLibrary import gaussian_fourier_transformed_spherical
from xframe.library.mathLibrary import spherical_to_cartesian

from .fxs_invariant_tools import harmonic_coeff_to_deg2_invariants_2d
from .fxs_invariant_tools import harmonic_coeff_to_deg2_invariants_3d
from . import fxs_invariant_tools as i_tools

from xframe import settings
from xframe import database
from xframe import Multiprocessing

log=logging.getLogger('root')

class RealProjection:
    # collection of possible real constraints
    # generation routine of a real_constraint function has to be named generate_<name>_projection
    # where <name> is the same string that is used in the settings file
    # each real support function takes as input an electron density (complex array)
    # and outputs a list [new_density,mask] where new_density is the changed density and mask is true at points in which the input denisity was modified.
    def __init__(self,opt,metadata):
        #self.initial_mask = False
        self.enforce_initial_support = True
        self.opt = opt
        self.real_grid = metadata['real_grid']
        self.auto_correlation = metadata.get('auto_correlation',False)
        self.metadata = metadata
        self._initial_mask = ~self.generate_initial_support_mask()
        self._mask = [self._initial_mask.copy()]
        self._random_mask = (np.random.rand(*self._initial_mask.shape)>0.9)
        self.projection = self.assemble_projection()
            
    
    @property
    def initial_support(self):
        return ~self._initial_mask.copy()


    @property
    def support(self):
        return ~self._mask[0]
    @support.setter
    def support(self,support):
        if self.enforce_initial_support:
            self._mask[0] = self._initial_mask | (~support)
        else:
            self._mask[0] = ~support        
    
    @property
    def mask(self):
        return self._mask[0]
    
    @mask.setter
    def mask(self,mask):
        if self.enforce_initial_support:
            self._mask[0] = self._initial_mask | mask
        else:
            self._mask[0] = mask
            
            
    def generate_support_projection(self):
        mask = self._mask
        def support_projection(density):
            #log.info('fu')
            #log.info('\n initial mask type = {} \n'.format(type(self.initial_mask)))
            #log.info(mask[0][:10,0])
            m= mask[0]
            density[m]=0
            return [density,m]
        return support_projection

    def generate_value_threshold_projection(self):
        threshold = self.opt['value_threshold'].get('threshold',0.0)
        threshold_projection = create_threshold_projection(threshold)
        return threshold_projection

    def generate_limit_imag_projection(self):        
        thresh = self.opt['limit_imag'].get('threshold',0.0)        
        def projection(density):
            imag_part = density.imag
            is_invalid_mask = np.abs(imag_part) >= thresh
            imag_part[is_invalid_mask] = 0
            return [density,is_invalid_mask]
        return projection
    def generate_average_center_projection(self):
        thresh = int(self.opt['average_center'].get('max_radial_id',1))
        dimension = self.real_grid.array.shape[-1]
        log.info('dim = {}'.format(dimension))
        axes = tuple(range(1,dimension))
        if dimension==2:
            def projection(density):
                density[:thresh]=np.mean(density[:thresh],axis=axes)[:,None]
                return [density,False]
        elif dimension==3:
            def projection(density):
                density[:thresh]=np.mean(density[:thresh],axis=axes)[:,None,None]
                return [density,False]
        return projection
    def assemble_projection(self):
        projections = []
        mask_dict = {}
        for key in self.opt.apply:
            try:
                projections.append([key,getattr(self,'generate_'+key+'_projection')()])
                mask_dict[key]=False
            except AttributeError as e:
                log.error('projection {} not known. Ignoring it.'.format(key))
        mask_dict['all']=False
        def real_projection(data):
            mask = False
            for name,projection in projections:
                #log.info(name)
                data,p_mask=projection(data)
                # individual masks are supposed to be True where projection changes the data
                mask_dict[name]=p_mask
                mask |= p_mask
                mask_dict['all']=mask
            return [data,mask_dict]
        return real_projection


    def generate_initial_support_mask(self):
        opt = self.opt.support.initial_support
        realGrid = self.real_grid
        support_type=opt['type']
        if support_type=='max_radius':
            maxR=opt['max_radius']
            #log.info(f'given_maximal radius is {maxR}')
            support_mask=np.where(realGrid[...,0]<maxR,True,False)
        elif support_type=='auto_correlation':
            assert isinstance(self.auto_correlation,np.ndarray), 'auto correlation not specified.'
            threshold=opt.auto_correlation.threshold
            max_value=np.max(self.auto_correlation)
            support_mask=self.auto_correlation >= threshold*max_value
            support_mask[realGrid[...,0]>settings.project.particle_radius]=0
        else:
            e=AssertionError('Initial support type "{}" is not known.'.format(support_type))
            log.error(e)
            raise e
        #    log.info('support mask[:]={}'.format(supportMask))
        #pres=heatPolar2D()
        #pres.present(auto_correlation.data.array,grid=realGrid)
        #pres.present(supportMask,grid=realGrid)
        return support_mask


def generate_initial_support_mask(opt,realGrid,projection):
    support_type=opt['type']
    if support_type=='max_radius':
        maxR=opt['max_r']
        supportMask=np.where(realGrid[...,0]<maxR,True,False)
    elif support_type=='auto_correlation':
        raise NotImplementedError()
        threshold=initial_support_specifier['threshold']        
        maxValue=np.max(auto_correlation.data)
        supportMask=auto_correlation.data>=(1-threshold)*maxValue
    else:
        e=AssertionError('Initial support type "{}" is not known.'.format(support_type))
        log.error(e)
        raise e
    #    log.info('support mask[:]={}'.format(supportMask))
    #pres.present(auto_correlation.data.array,grid=realGrid)
    #pres.present(supportMask,grid=realGrid)
    return ~supportMask


class ShrinkWrapParts():
    def __init__(self,real_grid,reciprocal_grid,initial_support,options = {}):# threshold = 0.04,gaussian_sigma = 2,mode = 'threshold',mode_options = {}):
        self.mode_routines = {'threshold':self.generate_get_new_mask_threshold,'fixed_volume':self.generate_get_new_mask_fixed_volume}

        mode = options.get('mode','threshold')
        mode_options = options.get(mode,{})
        
        dimension = real_grid[:].shape[-1]
        
        threshold = options.get('thresholds',[0.06])[0]
        #log.info(f'threshold = {threshold}')
        if dimension == 2:
            self.default_sigma = np.pi/(reciprocal_grid[:,0,0].max())
        elif dimension == 3:
            self.default_sigma = np.pi/(reciprocal_grid[:,0,0,0].max())
        gaussian_sigma = self.default_sigma
        #log.info(f'default sigma = {self.default_sigma}')
  
        self.mode = mode
        self.mode_options = mode_options
        self.real_grid = real_grid[:]
        self.reciprocal_grid = reciprocal_grid[:]
        dimension = self.real_grid.shape[-1]
        if dimension ==2:
            self.integrator = PolarIntegrator(self.real_grid)
        elif dimension == 3:
            self.integrator = SphericalIntegrator(self.real_grid)
        self.initial_support = initial_support
        self.initial_volume = self.integrator.integrate_normed(initial_support.astype(float))
        
        self._threshold = [threshold]
        self._gaussian_sigma = [gaussian_sigma]
        self.gaussian_values = gaussian_fourier_transformed_spherical(self.reciprocal_grid,self._gaussian_sigma[0])
        
        self.get_new_mask = self.mode_routines[self.mode]()
        self.multiply_with_ft_gaussian = self.generate_multiply_by_ft_gaussian()
        #log.info(f'\n Shrink wrap mode = {mode} , get new mask routine = {self.get_new_mask} \n')
    @property
    def threshold(self):
        return self._threshold[0]
    @threshold.setter
    def threshold(self,value):
        if value<0:
            self._threshold[0]=0
            log.warning('Shrikwrap threshold has to lie in [0,1] but given value is {}. Projecting threshold to {}.'.format(value,self._threshold[0]))
        elif value>=1:
            self._threshold[0]=1
            log.warning('Shrikwrap threshold has to lie in [0,1] but given value is {}. Projecting threshold to {}.'.format(value,self._threshold[0]))
        else:
            self._threshold[0]=value


    @property
    def gaussian_sigma(self):
        return self._gaussian_sigma[0]
    @gaussian_sigma.setter
    def gaussian_sigma(self,value):
        value_valid_type = not ( (not np.issubdtype(np.array(value).dtype,np.number)) or isinstance(value,bool))
        value_valid = False
        if value_valid_type:
            value_valid = value>0
        if value_valid:
            self._gaussian_sigma[0]=value
        else:
            self._gaussian_sigma[0]=self.default_sigma
        self.gaussian_values[:] = gaussian_fourier_transformed_spherical(self.reciprocal_grid,self._gaussian_sigma[0])
        
    def generate_get_new_mask_threshold(self):
        threshold = self._threshold
        def get_new_mask(convolution_data):
            #log.info(f'\n SW threshold = {threshold} and sigma = {self._gaussian_sigma}\n')
            convolution_data=convolution_data.real #np.abs(convolution_data).real
            #convolution_data=np.abs(convolution_data).real
            #convolution_data=np.abs(convolution_data).real #np.abs(convolution_data).real
            convolution_data[convolution_data<0]=0
            max_value= convolution_data.real.max()
            min_value= convolution_data.real.min()
            diff = max_value-min_value
            new_mask = convolution_data >= min_value + threshold[0]*diff
            return new_mask
        return get_new_mask
    
    def generate_get_new_mask_fixed_volume(self):
        try:
            target_volume = self.initial_volume*self.mode_options['volume'] # number in [0,1] indication volume fraction relative to self.initial_volume
        except KeyError as e:
            log.error('selected ShrinkWrap mode "fixed_volume" but no volume was specified in mode options {}'.format(self.mode_options))
            raise e
        integrate = self.integrator.integrate_normed
        old_volume = [self.initial_volume]
        d_vol_thresh = 0.2
        def get_new_mask(convolution_data):
            convolution_data=convolution_data.real
            max_value= convolution_data.real.max()
            min_value= convolution_data.real.min()
            diff = max_value-min_value
            def new_volume(threshold):
                new_mask = (convolution_data >= min_value + threshold*diff) & self.initial_support
                _vol = integrate(new_mask.astype(float))
                metric = abs(_vol-target_volume)
                rate_of_change = abs(old_volume[0] - _vol)/old_volume[0]
                #log.info('rate_of_change = {} threshold = {}'.format(rate_of_change,threshold))
                if rate_of_change>d_vol_thresh:
                    metric = np.inf
                return metric
            opti_result = minimize_scalar(new_volume,bounds=(0.0,1.0),method='golden')
            threshold = opti_result.x
            log.info('Optimization result = {}'.format(opti_result))
            new_mask = (convolution_data >= min_value + threshold*diff) & self.initial_support
            new_volume = integrate(new_mask.astype(float))
            log.info(f'old volume = {old_volume[0]} new volume = {new_volume}, target_volume={target_volume}')
            old_volume[0] = new_volume
            return new_mask
        return get_new_mask

    
    def generate_multiply_by_ft_gaussian(self):
        gaussian_values = self.gaussian_values
        def multiply_ft_gaussian(data):
            return data*gaussian_values
        return multiply_ft_gaussian

    
### FXS Projections
def generate_fxs_projection(specifier):
    dimension=specifier['fxs_data'].dimension
    if dimension==2:
        FXS_projection=construct_fxs_projection_parts_2D(specifier)
    else:
        raise NotImplementedError
    return FXS_projection

def calculate_fxs_projection_vector_old(fxs_data,positive_harmonic_orders):
    bCoefficients=fxs_data.bCoeff
    
    def firstEigenvectorAndValue(expansionCoefficient):
        matrix=bCoefficients.data.array[:,:,expansionCoefficient]
#        log.info('matrix of shape {} to take eigen values from = \n{}'.format(matrix.shape,matrix))
        eigValues,eigVectors=np.linalg.eigh(matrix)
        max_arg=eigValues.argmax()
        return [eigVectors[:,max_arg],eigValues[max_arg]]
    listOfEigenvectorsAndValues=np.array(list(map(firstEigenvectorAndValue,positive_harmonic_orders)),dtype=np.object)    
    eigenValues=listOfEigenvectorsAndValues[:,1].astype(complex)
    eigenVectors=np.array(tuple(listOfEigenvectorsAndValues[:,0]),dtype=complex)
    projection_vector=np.swapaxes(eigenVectors,0,1)*np.sqrt(eigenValues.astype(complex))
    fxs_data.projection_vector=projection_vector
    return fxs_data

def calculate_fxs_projection_vector(fxs_data,reciprocal_projection_specifier):
    positive_harmonic_orders=reciprocal_projection_specifier['positive_harmonic_orders']
    #log.info('projection orders={}'.format(positive_harmonic_orders))
    bCoefficients=fxs_data.bCoeff
    
    def firstEigenvectorAndValue(expansionCoefficient):
        matrix=bCoefficients[:,:,expansionCoefficient]
#        log.info('matrix of shape {} to take eigen values from = \n{}'.format(matrix.shape,matrix))
        eigValues,eigVectors=np.linalg.eigh(matrix)
        max_arg=eigValues.argmax()
        
        return [eigVectors[:,max_arg],eigValues[max_arg]]
    listOfEigenvectorsAndValues=np.array(list(map(firstEigenvectorAndValue,positive_harmonic_orders)),dtype=np.object)    
    eigenValues=listOfEigenvectorsAndValues[:,1].astype(complex)
    #log.info('lambdas={}'.format(eigenValues))
    eigenVectors=np.array(tuple(listOfEigenvectorsAndValues[:,0]),dtype=complex)
    projection_vector=np.swapaxes(eigenVectors,0,1)*np.sqrt(eigenValues.astype(complex))    
    fxs_data.projection_vector=projection_vector
#    log.info('shape ={} projection_vector={}'.format(projection_vector.shape[1],np.abs(np.sum(projection_vector,axis=0))))
    return fxs_data


def construct_fxs_projection_parts_2D(specifier):
    fxs_data=specifier['fxs_data']
    specifier['projection_vector']=modify_projection_vector_2D(fxs_data.projection_vector,specifier)
    mask=specifier.get('mask',False)
    if isinstance(mask,bool):
        specifier['mask']=True
    
    use_SO_freedom=specifier['SO_freedom']['use']
    
    fxs_projection=generate_coefficient_projection_2D(specifier)
    approximate_unknowns_without_SO=generate_approximate_unknowns_2D(specifier)

    if use_SO_freedom:
        apply_SO_freedom=generate_apply_SO_freedom_2D(specifier)
        def approximate_unknowns(intensity_harmonic_coefficients):
            return apply_SO_freedom(approximate_unknowns_without_SO(intensity_harmonic_coefficients))
    else:
        approximate_unknowns=approximate_unknowns_without_SO
            
    projection_dict={'projection':fxs_projection,'approx_unknowns':approximate_unknowns,'proj_mask':mask}
    return projection_dict
        
def modify_projection_vector_2D(projection_vector,specifier):
    rescale_projection=specifier.get('rescale_projection_to_1',False)
    average_intensity=specifier['fxs_data'].aInt.data.astype(complex)
    positive_harmonic_orders=specifier['positive_harmonic_orders']
    odd_order_mask=positive_harmonic_orders%2==1
    vector=projection_vector.copy()
    vector[:,odd_order_mask]=0
    vector[:,0]=average_intensity
    if rescale_projection:        
        max_vector_value=np.abs(np.max(vector))
        vector/=max_vector_value
#    log.info('projection vector for SO={}'.format(np.abs(np.sum(projection_vector,axis=0))))
    return vector

def generate_approximate_unknowns_2D(specifier):
    positive_harmonic_orders=specifier['positive_harmonic_orders']
    projection_vector=specifier['projection_vector']
    reciprocal_grid=specifier['reciprocal_grid']
 #   log.info(np.conjugate(np.sum(projection_vector,axis=0)))
    def approximate_unknowns(intensity_harmonic_coefficients):
        scalar_prod_Im_vm_summands=intensity_harmonic_coefficients[:,positive_harmonic_orders]*np.conjugate(projection_vector)*reciprocal_grid[:,positive_harmonic_orders,0]
        scalar_prod_Im_vm=np.sum(scalar_prod_Im_vm_summands,axis=0)
#        log.info(scalar_prod_Im_vm)
        unknowns=scalar_prod_Im_vm/np.where(np.abs(scalar_prod_Im_vm)==0,1,np.abs(scalar_prod_Im_vm))
        unknowns[0]=1
        #log.info('unknowns={}'.format(unknowns))
        return unknowns
    return approximate_unknowns

def generate_coefficient_projection_2D(specifier):
    positive_harmonic_orders=specifier['positive_harmonic_orders']
    projection_vector=specifier['projection_vector']
    mask=specifier['mask']
    copy = np.array
    def fxs_projection(intensity_harmonicCoefficients,unknowns):
        projected_intensity_coefficients_array=copy(intensity_harmonicCoefficients)
        projected_intensity_coefficients_array[:,positive_harmonic_orders]=projection_vector*unknowns
        #apply mask
        intensity_harmonic_coefficients=np.where(mask,projected_intensity_coefficients_array,intensity_harmonicCoefficients).astype(complex)
        return intensity_harmonic_coefficients
    return fxs_projection


def generate_orientation_matching_2D(n_positive_harmonic_orders):
    previous_phases=[False]
    previous_mask=[True]
    def orientation_matching(intensity_harmonic_coefficients,unknowns):
        non_zero_mask=unknowns!=0
        mask=previous_mask[0]*non_zero_mask
        phases=np.zeros(unknowns.shape)
        phases[non_zero_mask]=(-1.j*np.log(unknowns[non_zero_mask])).real
        if not isinstance(previous_phases[0],bool):
            non_zero_phases=phases[mask]
            optimal_phase=np.sum(non_zero_phases-previous_phases[0][mask])/len(non_zero_phases)
            shift=np.exp(-1.j*np.arange(n_positive_harmonic_orders)*optimal_phase)
#            log.info(shift)
            intensity_harmonic_coefficients*=shift
        previous_phases[0]=phases
        previous_mask[0]=non_zero_mask
        return intensity_harmonic_coefficients
    return orientation_matching

def generate_orientation_matching(dimensions,n_positive_harmonic_orders):
    if dimensions == 2:
        orientation_matching=generate_orientation_matching_2D(n_positive_harmonic_orders)
    elif dimensions == 3:
        def orientation_matching(intensity_harmonic_coefficients,unknowns):
            raise NotImplementedError()
    else:
        log.error('dimensions must be 2 or 3 but {} was given'.format(dimensions))
    return orientation_matching


class ReciprocalProjection:
    def load_data(self,data):
        opt = settings.project
        self.dimensions = data['dimensions']
        self.xray_wavelength = data['xray_wavelength']
        self.average_intensity = data['average_intensity']
        #log.info('aint type ={}'.format(type(self.average_intensity)))
        self.data_radial_points = data['data_radial_points'][:]
        self.data_angular_points = data['data_angular_points'][:]
        self.data_max_q = np.max(self.data_radial_points)
        self.data_min_q = np.min(self.data_radial_points)
        #self.pi_in_q = data.get('pi_in_q',False)
        #self.pi_in_q = data['pi_in_q']
        #log.info(f'data q_min = {self.data_min_q} max_q ={self.data_max_q}')
        
        self.data_max_order=data['max_order']
        self.data_projection_matrices = data['data_projection_matrices']
        self.data_low_resolution_intensity_coefficients = data.get('data_low_resolution_intensity_coefficients',self.data_projection_matrices)
        #log.info(f' pr matrix type {type(self.data_projection_matrices)}')
        self.data_projection_matrices_q_id_limits = data.get('data_projection_matrices_q_id_limits',False)
        #log.info(f'data proj matrices len = {len(self.data_projection_matrices)}')
        #if not isinstance(self.data_projection_matrices,np.ndarray):
        #    tmp = np.empty(len(self.data_projection_matrices),object)
        #    for i,pm in enumerate(self.data_projection_matrices):
        #        tmp[i]=pm
        #    self.data_projection_matrices = tmp

                
    def __init__(self,grid,data,max_order):
        self.load_data(data)
        if self.dimensions==2:
            self.integrated_intensity = midpoint_rule(self.average_intensity.data * self.data_radial_points , self.data_radial_points,axis = 0)*2*np.sqrt(np.pi)
        else:
            self.integrated_intensity = midpoint_rule(self.average_intensity.data * self.data_radial_points**2 , self.data_radial_points,axis = 0)*2*np.sqrt(np.pi)
        opt = settings.project.projections.reciprocal        
        self.opt = opt
        self.grid=grid        
        self.radial_points=grid.__getitem__((slice(None),)+(0,)*self.dimensions) # expression is just grid[:,0,0,0] in 3D case ans grid[:,0,0] in 2D
        self.max_q = np.max(self.radial_points)
        #self.input_is_complex=opt.input_is_halfcomplex
        
        self.positive_orders = np.arange(max_order+1)        
        
        self.projection_orders = self.read_projection_orders(opt.get('used_orders',opt.used_order_ids),max_order)
        #log.info('projection_orders = {}'.format(self.projection_orders))
        self.projection_order_ids_local = np.arange(len(self.projection_orders))
        self.projection_order_ids_global = np.isin(self.positive_orders,self.projection_orders).nonzero()[0]
                    
        self.used_order_ids = opt.used_order_ids
        self.used_orders = {order:id for (order,id) in zip(self.positive_orders,self.used_order_ids)}
        self.use_SO_freedom=opt.SO_freedom.use
        
        # Won't work without the list around opt.number_of_particles.initial .
        # This way number_of_particles[0] automatically contains updated versions of the particle number.  
        self.number_of_particles = [opt.number_of_particles.initial] 
        self.number_of_particles_dict = {'number_of_particles':self.number_of_particles,'negative_fraction':[],'gradient':[]}

        #log.info('data_projection_matrices shape = {}'.format(tuple(d.shape for d in self.data_projection_matrices)))
        pm,low_res = self._regrid_data()        
        #log.info(f'len regridded proj mat = {len(pm)}')
        
        self.projection_matrices = pm
        nq = self.grid.shape[0]
        if self.dimensions ==3:
            self.full_projection_matrices = [np.zeros((nq,min(nq,2*o+1)),dtype=complex) for o in range(max_order+1)]
        elif self.dimensions == 2:
            self.full_projection_matrices = np.zeros((max_order+1,nq),dtype=complex)
        for oid,pm in zip(self.used_order_ids,self.projection_matrices):
            self.full_projection_matrices[oid]=pm
            
        self.low_resolution_intensity_coefficients = low_res
        #log.info('projection_matrices after regrid = {}'.format(tuple(d.shape for d in self.projection_matrices)))
        self.projection_matrices=self.modify_projection_matrices(opt)
        #log.info('projection_matrices after modify = {}'.format(tuple(d.shape for d in self.projection_matrices)))
        self.assert_projection_matrices_have_right_shape()        
        self.radial_mask=self.generate_radial_mask(opt.q_mask)        
        self._fixed_intensity = [np.ones(grid.shape[:-1],dtype=complex)]        
        mtip_projection_base=self.generate_coeff_projection_base()        
        self.mtip_projection_base  = mtip_projection_base        
        self.mtip_projection = self.generate_coeff_projection(mtip_projection_base)            
        self.approximate_unknowns=self.generate_approximate_unknowns()
            
        #if opt.number_of_particles.GPU:            
        #    self.particle_number_projection = self.generate_number_of_particles_porjection_gpu()
        #else:
        #    self.particle_number_projection = self.generate_number_of_particles_porjection()
        
        self.project_to_modified_intensity = self.generate_project_to_modified_intensity()
        self.project_to_fixed_intensity = self.generate_project_to_fixed_intensity()
        self.deg2_invariants = self.calc_deg2_invariants()

        self.remaining_SO_projection=False
        if self.use_SO_freedom:
            radial_high_pass=opt.SO_freedom.get('radial_high_pass',0.2)
            self.remaining_SO_projection = self.generate_remaining_SO_projection(radial_high_pass=radial_high_pass)

    

    @property
    def fixed_intensity(self):
        return self._fixed_intinsity[0]
    @fixed_intensity.setter
    def fixed_intensity(self,value):
        self._fixed_intensity[0]=value.real
        
    def read_projection_orders(self,projection_orders,max_harmonic_order):
        try:
            assert projection_orders.max()<=self.data_max_order, 'Max available harmonic order from dataset is {} but the maximal projection order is {}. Restricting ordes to the ones present in the dataset.'.format(self.data_max_order,projection_orders.max())
        except AssertionError as e:
            log.warning(e)
            projection_orders = projection_orders[projection_orders<=self.data_max_order]
        try:
            assert projection_orders.max()<=max_harmonic_order, 'The maximal harmonic oder given by the input harmonic analysis {} is smaller than the specified maximal projection order {}.Limiting projection orders and continue.'.format(max_harmonic_order,projection_orders.max())
        except AssertionError as e :
            log.warning(e)
            projection_orders = projection_orders[projection_orders<=max_harmonic_order]
        return projection_orders
        
    def assert_projection_matrices_have_right_shape(self):
        try:
            n_radial_points_grid = len(self.radial_points)
            n_radial_points_projection_matrix = self.projection_matrices[0].shape[0]
            assert n_radial_points_grid == n_radial_points_projection_matrix,'Mismatch between th number of radial sampling points in reconstruction grid {}  and projection data {}. Abort reconstruction !'.format(n_radial_points_grid,n_radial_points_projection_matrix)
        except AssertionError as e:
            log.error(e)
            raise e
            
    def extract_used_options(self):
        fourier_opt = settings.project.fourier_transform
        rp_opt = settings.project.projections.reciprocal
        opt = DictNamespace(
            pos_orders = fourier_opt.pos_orders,
            **rp_opt
        )
        return opt
    def generate_radial_mask(self,mask_opt):
        '''
        Constructs the radial mask (mask of momentum transfair values). Masked values are False and will not be used 
        in the reciprocal coefficient projection defined in self.generate_coeff_projection.
        '''
        radial_points = self.radial_points
        mask = np.full((len(self.positive_orders),len(radial_points)),False)        
        if ~isinstance(self.data_min_q,bool):
            data_mask =  mask | (radial_points>=self.data_min_q) & (radial_points<=self.data_max_q)
        
        if isinstance(mask_opt,(dict,DictNamespace)):
            mtype=mask_opt['type']
            #log.info('mask type = {}'.format(mtype))
            if mtype=='none':
                mask = True
            elif mtype == 'from_projection_matrices':
                data_limit_ids = self.data_projection_matrices_q_id_limits['I1I1']
                for mask_part,lim in zip(mask,data_limit_ids):
                    min_q = self.data_radial_points[lim[0]]
                    max_q = self.data_radial_points[lim[1]-1]
                    mask_part[:] = (radial_points>min_q) & (radial_points<max_q)
            elif mtype == 'manual':
                manual_opt = mask_opt['manual']
                manual_type = manual_opt['type']
                #log.info('manual type = {}'.format(manual_type))
                if manual_type == 'region':
                    region=manual_opt['region']
                    radial_points=self.radial_points
                    #log.info(f'region = {region}')
                    if (region[0] == False) and (region[1] != False):
                        #log.info(f'radial_points = {radial_points}')
                        mask[:]=(radial_points<region[1])[None,:]
                    elif (region[0] != False) and (region[1] == False):                    
                        mask[:]=(radial_points>=region[0])[None,:]
                        log.info('radial mask non zero count {} of {}'.format(np.sum(mask),np.prod(mask.shape)))
                    elif (region[0] != False) and (region[1] != False):
                        mask[:]=((radial_points>=region[0])  & (radial_points<region[1]))[None,:]
                    else:
                        mask[:]=True
                        log.info('nothing masked  {}'.format(mask.all()))
                elif manual_type == 'order_dependent_line':
                    points = manual_opt['order_dependent_line']
                    log.info('points = {}'.format(points))
                    orders = self.positive_orders
                    data_grid = GridFactory.construct_grid('uniform',[orders,radial_points])                 
                    mask = (-1*distance_from_line_2d(np.array(points),data_grid[:]))>=0
        else:
            log.warning('Could not parse projections.reciprocal.q_mask option. Proceeding without custom q_mask')
            mask = True
                
        mask = mask & data_mask
        return mask

    def calc_deg2_invariants(self):
        if self.dimensions == 2:
            invariants = harmonic_coeff_to_deg2_invariants_2d(self.projection_matrices.T)
        elif self.dimensions == 3:
            invariants = harmonic_coeff_to_deg2_invariants_3d(self.projection_matrices)
        #log.info('rp deg 2 invariant shape = {}'.format(invariants.shape))
        return invariants

    def _regrid_data(self):
        dim = self.dimensions
        order_ids=list(self.used_orders.values())
        have_same_shapes = self.radial_points.shape == self.data_radial_points.shape
        projection_matrices = self.data_projection_matrices[order_ids]
        needs_regridding = True
        interpolation_type = self.opt.regrid.interpolation
        if have_same_shapes:
            if (self.data_radial_points == self.radial_points).all():
                needs_regridding = False
        #log.info('needs regridding = {}'.format(needs_regridding))
        #log.info('initial projection matrix shape = {}'.format(projection_matrices[-1].shape))   
        if needs_regridding:
            r_pt=NestedArray(self.radial_points[:,None],1)
            #log.info('r_pt = {}'.format(r_pt[:10]))
            data_r_pt=NestedArray(self.data_radial_points[:,None],1)
            #log.info('n new points={} n old points ={}'.format(len(r_pt[:]),len(data_r_pt[:])))
            low_res = False
            if dim == 2:                
                self.average_intensity.regrid(r_pt,options={'apply_over_axis':0,'fill_value': 0.0,'interpolation':interpolation_type})
                #log.info('proj data shape = {}'.format(self.data_projection_matrices.shape))
                #log.info(f'order ids = {order_ids}')
                projection_matrices = ReGrider.regrid(np.array(self.data_projection_matrices)[order_ids,...],data_r_pt,'cartesian',r_pt,'cartesian',options={'apply_over_axis':0,'fill_value': 0.0,'interpolation':interpolation_type})
                #log.info("data_r_min = {} grid r min = {}".format(data_r_pt.array.min(),r_pt.array.min()))
                #log.info('regrided projection matrices complete shape = {}'.format(projection_matrices.shape))
                #log.info('regrided projection matrices shape = {}'.format(projection_matrices.shape))
                #self.average_intensity.regrid(r_pt)
                #projection_matrices = ReGrider.regrid(self.data_projection_matrices[...,order_ids],data_r_pt,'cartesian',r_pt,'cartesian',options={'apply_over_axis':1,'fill_value': 0.0,'interpolation':interpolation_type})
            elif dim == 3:
                self.average_intensity.regrid(r_pt,options={'apply_over_axis':0,'fill_value': 0.0,'interpolation':interpolation_type})
                data_projection_matrices=self.data_projection_matrices
                projection_matrices = tuple(ReGrider.regrid(data_projection_matrices[o_id],data_r_pt,'cartesian',r_pt,'cartesian',options={'apply_over_axis':1,'fill_value': 0.0,'interpolation':interpolation_type}) for o_id in order_ids)
                if isinstance(self.data_low_resolution_intensity_coefficients,np.ndarray):
                    data_low_res = self.data_low_resolution_intensity_coefficients
                    low_res = tuple(ReGrider.regrid(data_low_res[o_id],data_r_pt,'cartesian',r_pt,'cartesian',options={'apply_over_axis':1,'fill_value': 0.0,'interpolation':interpolation_type}) for o_id in np.arange(len(data_low_res)))
                    
        #log.info('regrided projection matrices shape = {}'.format(projection_matrices[-1].shape))
        return projection_matrices,low_res
    

    def modify_projection_matrices(self,opt):
        dim = self.dimensions
        use_averaged_intensity=opt.get('use_averaged_intensity',False)
        use_odd_orders_to_0=opt.get('odd_orders_to_0',False)
        
        average_intensity=self.average_intensity.data.astype(complex)
        used_orders=self.used_orders
        odd_order_mask=np.array(tuple(used_orders))%2==1
        if dim == 2:
            proj_matrices=self.projection_matrices.copy()
        elif dim == 3:
            proj_matrices=[matrix.copy() for matrix in self.projection_matrices]
        #log.info(f'use average intensity = {use_averaged_intensity}')
        
        if use_odd_orders_to_0:
            if dim == 2:        
                proj_matrices[odd_order_mask,:]=0
            elif dim == 3:
                for odd_order in np.array(tuple(used_orders))[odd_order_mask]:
                    proj_matrices[used_orders[odd_order]][:]=0

        if use_averaged_intensity:
            zero_id = used_orders[0]            
            if dim == 2:
                #log.info('proj dtype = {} average dtype = {}'.format(proj_matrices.dtype,average_intensity.dtype))
                proj_matrices[zero_id] = average_intensity
                #proj_matrices[:,used_orders[0]]=average_intensity
            elif dim == 3:
                # assumes schmidt seminormalized definition of spherical harmonics -> factor of 2*np.sqrt(pi) 
                proj_matrices[zero_id]= average_intensity[:,None].real*2*np.sqrt(np.pi)
                
        # internally orthonarmalized spherical harmonics are used but data is ussually supplied for schmidt seminormalized spherical harmonics
        if dim == 3:
            for pm in proj_matrices:
                pm[:]*=2  
        return proj_matrices


    def generate_approximate_unknowns(self):
        dim = self.dimensions
        used_orders=self.used_orders
        order_ids=tuple(used_orders.values())
        proj_matrices=self.projection_matrices
        radial_points=self.radial_points
        if dim == 2:
            proj_matrices = self.projection_matrices.T
            sum=np.sum
            where=np.where
            abs=np.abs
            conjugate=np.conjugate
            unknowns = np.zeros(len(order_ids),dtype = complex)
            def approximate_unknowns(intensity_harmonic_coefficients):
                #log.info(proj_matrices.shape)
                scalar_prod_Im_vm_summands=intensity_harmonic_coefficients[:,order_ids]*conjugate(proj_matrices)*radial_points[:,None]
                scalar_prod_Im_vm=sum(scalar_prod_Im_vm_summands,axis=0)
                non_zero_mask = scalar_prod_Im_vm !=0
                unknowns[non_zero_mask] = scalar_prod_Im_vm[non_zero_mask]/abs(scalar_prod_Im_vm[non_zero_mask])
                #        log.info(scalar_prod_Im_vm)
                unknowns[~non_zero_mask]=1
                #tmp = np.concatenate((unknowns,unknowns[:0:-1].conj()))
                #unknowns[0]=1
                #log.info('unknowns={}'.format(unknowns))
                return unknowns#np.zeros_like(unknowns)
            
            if self.use_SO_freedom:
                SO_order_id =self.get_SO_application_order()
                def function(intensity_harmonic_coefficients):
                    unknowns = approximate_unknowns(intensity_harmonic_coefficients)
                    unknowns[SO_order_id]=1
                    return unknowns
            else:
                function = approximate_unknowns
            
        if dim == 3:
            D=np.diag(radial_points) #diagonal matrix of radial points
            PDs=tuple(proj_matrices[_id].T.conj() @ D**2 for _id in order_ids)
            matmul=np.matmul
            svd=np.linalg.svd
            determinant = np.linalg.det
            loge = np.log
            n_orders = len(PDs)
            unknowns = tuple(np.zeros((len(PD),2*o+1),dtype = complex) for PD,o in zip(PDs,used_orders))
            def approximate_unknowns(intensity_harmonic_coefficients):
                #log.info('len harmonic coeff = {}'.format(len(intensity_harmonic_coefficients)))
                for unknown,PD,oid in zip(unknowns,PDs,order_ids):
                    I = intensity_harmonic_coefficients[oid]
                    matmul(*svd(PD @ I,full_matrices=False)[::2],out = unknown)  # PD @ Intensity is  B^\dagger A in a Procrustres Problem min|A-BR|
                #log.info('unknowns shape ={}'.format(unknowns[-1].shape))
                return unknowns
            if self.use_SO_freedom:
                radial_high_pass = self.opt.SO_freedom.radial_high_pass
                ranked_ids,ranked_orders,_ = i_tools.rank_projection_matrices(self.dimensions,proj_matrices,self.positive_orders,radial_points,radial_high_pass = radial_high_pass)
                SO_order_id = ranked_ids[0]
                SO_order = self.positive_orders[SO_order_id]
                max_order = self.positive_orders.max()
                ms_per_order = []
                for order in np.arange(max_order+1):
                    ms_per_order.append(np.concatenate((np.arange(order+1),np.arange(-order,0))))
                
                def function(intensity_harmonic_coefficients):
                    for ms,unknown,PD,I in zip(ms_per_order,unknowns,PDs,intensity_harmonic_coefficients):
                        u,s,vh = svd(PD @ I,full_matrices=False)
                        matmul(u,vh,out = unknown)  # PD @ Intensity is  B^\dagger A in a Procrustres Problem min|A-BR|
                    
                    u_SO = unknowns[SO_order_id]
                    u_SO[4,2] = u_SO[4,2].real                    #u_SO[4,2] = u_SO[4,2].real
                    #log.info(f'unknown SO_order imag part = {unknowns[SO_order_id][4,2].imag}')
                    return unknowns
            else:                
                function = approximate_unknowns
            
        return function
    
    def generate_coeff_projection_base(self):
        dim = self.dimensions
        projection_matrices=self.projection_matrices
        used_orders=self.used_orders
        zero_id  = used_orders.get(0,False)
        zero_pos = np.argmax(np.array(tuple(used_orders.keys())) == 0)
        
        use_averaged_intensity=self.opt.get('use_averaged_intensity',False)
        #average_intensity = self.average_intensity.data[:,None]
        #log.info('projection_matrices type={}'.format(type(projection_matrices)))
        #mask=self.mask_2d
        radial_mask = self.radial_mask
        #projection_matrices[zero_id]= np.abs(projection_matrices[zero_id])
        if dim == 2:
            where=np.where
            order_array=np.array(list(used_orders.values()))
            #order_array=np.concatenate((order_array,order_array[:0:-1]))
            new_intensity_coefficients=np.zeros((self.grid.shape[0],(self.grid.shape[1]+1)//2),dtype=complex)
            mult=np.multiply
            mask = np.zeros((self.grid.shape[0],(self.grid.shape[1]+1)//2),dtype=bool)
            for o_id in order_array:
                mask[:,o_id]=radial_mask[o_id]
            radial_mask2 = radial_mask[order_array].T
            #log.info('2d mask .shape = {}, 2d_radial_mask shape = {}, grid shape = {} len orders = {} \n\n\n'.format(mask.shape,radial_mask2.shape,self.grid.shape,order_array.shape))
            #projection_matrices=np.concatenate((projection_matrices,projection_matrices[:0:-1].conj()),axis=0)
            projection_matrices=projection_matrices.T
            
            qs = self.grid[:,0,0]
            sphere = spherical_formfactor(qs,radius = 10)
            if not isinstance(zero_id,bool):
                def mtip_projection(intensity_harmonic_coefficients,unknowns):
                    new_intensity_coefficients[:]=intensity_harmonic_coefficients
                    new_intensity_coefficients[mask] = (projection_matrices*unknowns[None,:])[radial_mask2]
                    new_intensity_coefficients[radial_mask[zero_id],zero_id] = projection_matrices[radial_mask[zero_id],zero_pos]
                    return new_intensity_coefficients                
            else:
                def mtip_projection(intensity_harmonic_coefficients,unknowns):
                    new_intensity_coefficients[:]=intensity_harmonic_coefficients
                    new_intensity_coefficients[mask] = (projection_matrices*unknowns[None,:])[radial_mask2]
                    return new_intensity_coefficients
        elif dim == 3:
            copy=np.array #array is faster than copy
            if not isinstance(zero_id,bool):
                def mtip_projection(intensity_harmonic_coefficients,unknowns):
                    projected_intensity_coefficients=[copy(coeff) for coeff in intensity_harmonic_coefficients]
                    for o_id in self.used_orders.values():
                        tmp_coeff = projection_matrices[o_id] @ unknowns[o_id]
                        projected_intensity_coefficients[o_id][radial_mask[o_id],...] = tmp_coeff[radial_mask[o_id],...]
                    projected_intensity_coefficients[zero_id][radial_mask[zero_id],...] = projection_matrices[zero_id][radial_mask[zero_id],...]
                    return projected_intensity_coefficients
            else:
                def mtip_projection(intensity_harmonic_coefficients,unknowns):
                    projected_intensity_coefficients=[copy(coeff) for coeff in intensity_harmonic_coefficients]
                    for o_id in self.used_orders.values():
                        tmp_coeff = projection_matrices[o_id] @ unknowns[o_id]
                        projected_intensity_coefficients[o_id][radial_mask[o_id],...] = tmp_coeff[radial_mask[o_id],...]
                    return projected_intensity_coefficients
        return mtip_projection
    
    def generate_coeff_projection(self,coeff_projection):
        dim = self.dimensions
        used_orders=self.used_orders
        zero_id = used_orders.get(0,'')
        radial_mask = self.radial_mask
        number_of_particles = self.number_of_particles
        if dim == 2:
            def fixed_projection(intensity_harmonic_coefficients,unknowns):
                projected_intensity_coefficients = coeff_projection(intensity_harmonic_coefficients,unknowns)
                #projected_intensity_coefficients[zero_id][radial_mask[zero_id]]/=np.sqrt(number_of_particles[0])
                #log.info(f'number of particles scaling factor = {1/np.sqrt(number_of_particles[0])}')
                #log.info(f'out_array = {projected_intensity_coefficients[:,zero_id].shape} in array = {self.projection_matrices[zero_id].shape}')
                projected_intensity_coefficients[:,zero_id]/=np.sqrt(number_of_particles[0])
                return projected_intensity_coefficients
        elif dim == 3:
            def fixed_projection(intensity_harmonic_coefficients,unknowns):
                projected_intensity_coefficients = coeff_projection(intensity_harmonic_coefficients,unknowns)
                #projected_intensity_coefficients[zero_id][radial_mask[zero_id]]/=np.sqrt(number_of_particles[0])
                #log.info(f'number of particles scaling factor = {1/np.sqrt(number_of_particles[0])}')
                projected_intensity_coefficients[zero_id][:]/=np.sqrt(number_of_particles[0])
                return projected_intensity_coefficients
        return fixed_projection

    def generate_project_to_modified_intensity(self):
        if settings.general.cache_aware:
            L2_cache = settings.general.L2_cache
            func = self._generate_project_to_modified_intensity_cache_aware(L2_cache)
        else:
            func = self._generate_project_to_modified_intensity_default()
        return func
    
    def generate_project_to_fixed_intensity(self):
        if settings.general.cache_aware:
            L2_cache = settings.general.L2_cache
            func = self._generate_project_to_modified_intensity_cache_aware(L2_cache,use_fixed_intensity=True)
        else:
            func = self._generate_project_to_modified_intensity_default(use_fixed_intensity=True)
        return func
    def _generate_project_to_modified_intensity_default(self,use_fixed_intensity=False):
        zeros = np.zeros
        density_shape = self.grid[:].shape[:-1]
        sqrt = np.sqrt
        mult = np.multiply
        nabs = np.abs
        new_reciprocal_density = np.zeros(density_shape,dtype =complex)
        intensity_multipliers = np.zeros(new_reciprocal_density.shape , dtype= complex)
        temp = np.zeros(new_reciprocal_density.shape , dtype= float)
        if not use_fixed_intensity:
            def project_to_modified_intensity(reciprocal_density,square,new_intensity):            
                non_zero_mask = (square>=0)  & (new_intensity.real>=0)            
                #log.info('square dtype = {}'.format(square.dtype))
                temp[non_zero_mask] = new_intensity.real[non_zero_mask]/square[non_zero_mask].real
                np.sqrt(temp,out = intensity_multipliers)
                intensity_multipliers[~non_zero_mask] = 0            
                mult(reciprocal_density ,intensity_multipliers,out = new_reciprocal_density)
                #log.info("old intesity sum = {} new intensity sum = {}".format(np.sum(square),np.sum(new_intensity)))
                #log.info('nans = {} infs = {}'.format(np.isnan(new_reciprocal_density),np.isinf(new_reciprocal_density)))
                #new_reciprocal_density = reciprocal_density * sqrt(new_intensity.real/(reciprocal_density*reciprocal_density.conj()).real)            
                return new_reciprocal_density
        else:
            new_intensity2 = self._fixed_intensity
            def project_to_modified_intensity(reciprocal_density,square):
                #log.info(new_intensity2[0][:10,0,0])
                non_zero_mask = (square>=0) & (new_intensity2[0]>=0)            
                #log.info('square dtype = {}'.format(square.dtype))
                temp.real[non_zero_mask] = new_intensity2[0][non_zero_mask]/square[non_zero_mask].real
                np.sqrt(temp,out = intensity_multipliers)
                intensity_multipliers[~non_zero_mask] = 0            
                mult(reciprocal_density ,intensity_multipliers,out = new_reciprocal_density)
                #log.info("old intesity sum = {} new intensity sum = {}".format(np.sum(square),np.sum(new_intensity)))
                #log.info('nans = {} infs = {}'.format(np.isnan(new_reciprocal_density),np.isinf(new_reciprocal_density)))
                #new_reciprocal_density = reciprocal_density * sqrt(new_intensity.real/(reciprocal_density*reciprocal_density.conj()).real)            
                return new_reciprocal_density
        
        return project_to_modified_intensity
    def _generate_project_to_modified_intensity_cache_aware(self,L2_cache,use_fixed_intensity=False):
        data_shape = self.grid[:].shape[:-1]
        #split_id,step = get_L2_cache_split_parameters(data_shape,np.dtype(complex),L2_cache)
        return self._generate_project_to_modified_intensity_default(use_fixed_intensity = use_fixed_intensity)

    #scaled version to donatellies n particle determination

    def rank_projection_matrix_orders_2d(self):
        radial_high_pass=self.opt.SO_freedom.get('radial_high_pass',0.2)
        radial_high_pass_index = int((len(self.radial_points)-1)*radial_high_pass)
        radial_high_pass=0
        if not isinstance(radial_high_pass_index,bool):
            radial_high_pass=radial_high_pass_index
        radial_points=self.radial_points[radial_high_pass:]


        orders = self.projection_orders
        #log.info(orders)
        order_ids= self.projection_order_ids_local
        even_order_mask= orders%2 == 0
        non_zero_mask= orders != 0
        order_mask=even_order_mask*non_zero_mask
        relevant_orders=orders[order_mask]
        #relevant_order_ids = order_ids[order_mask]

        projection_vector=self.projection_matrices[order_mask,radial_high_pass_index:].T        
        

        weighted_vect=projection_vector*radial_points[:,None]#/(np.concatenate(([100.0,100.0,6,8],harmonic_orders[4:])))[None,:]
        #log.info('weighted_vect shape =\n {}'.format(weighted_vect.shape))
        metric = np.mean(np.abs(weighted_vect),axis=0)#/orders[order_mask]
        #log.info(metric)
        sorted_indices = np.argsort(metric)[::-1]
        #log.info(sorted_indice)
        SO_order_indices=order_mask.nonzero()[0][sorted_indices]
        SO_orders=orders[SO_order_indices]
        return SO_order_indices,SO_orders,sorted_indices

    def get_SO_application_order(self):
        dim=self.dimensions
        if dim ==2 :
            SO_order_indices,SO_orders,sorted_indices=self.rank_projection_matrix_orders_2d()            
            #log.info('Selected SO_order_ids = {}'.format(SO_order_indices))
        elif dim == 3:
            raise NotImplementedError
        return SO_order_indices[0]
            
    def generate_apply_SO_freedom_2D(self,opt):
        dim=self.dimensions
        if dim ==2 :
            ordes_dict=self.used_orders
            orders=np.array(tuple(orders_dict))
            order_ids=tuple(orders_dict.values())
            
            radial_high_pass=opt['SO_freedom'].get('radial_high_pass',0.2)
            radial_high_pass_index = int((len(self.radial_points)-1)*radial_high_pass)
            radial_points=self.radial_points

            even_order_mask= orders%2 == 0
            if use_averaged_intensity:
                non_zero_mask= orders != 0
            else:
                non_zero_mask=True
            order_mask=even_order_mask*non_zero_mask
            relevant_orders=orders[order_mask]
            max_order=np.max(relevant_orders)
        
            projection_vector=self.projection_matrices[radial_high_pass_index:,order_mask]
            
            #pres_cart.present(np.swapaxes(np.abs(projection_vector),0,1),scale='log')
            #pres_1d.present(np.sum(np.abs(projection_vector),axis=0))
            sorted_projection_indices=self.rank_orders(projection_vector,radial_points[radial_high_pass_index:],relevant_orders)
            #log.info('sorted_projection_indices={}'.format(sorted_projection_indices))
            SO_order_index=order_mask.nonzero()[0][sorted_projection_indices[0]]
            #log.info('nonzero = {}'.format(order_mask.nonzero()[0]))
            log.info('SO order Index={}'.format(SO_order_index))
            #   log.info('first_order_index={}'.format(first_order_index))
            #   log.info('first_order={}'.format(first_order))
            
            def apply_SO_freedom(unknowns):
                #log.info('unknowns shape={}'.format(unknowns.shape))
                unknowns[SO_order_index]=1
                return unknowns
        elif dim == 3:
            raise NotImplementedError
        return apply_SO_freedom

    def generate_remaining_SO_projection(self,radial_high_pass=0.2):
        dim=self.dimensions
        if dim == 2:
            remaining_SO_projection=self.generate_remaining_SO_projection_2D(radial_high_pass=radial_high_pass)
        elif dim == 3:
            def remaining_SO_projection(harmonic_coefficients,fxs_unknowns):            
                raise NotImplementedError()        
        return remaining_SO_projection
    
    def generate_remaining_SO_projection_2D(self,radial_high_pass=0.2):
        radial_high_pass_index = int((len(self.radial_points)-1)*radial_high_pass)
        projection_vectors = self.projection_matrices
        n_angular_points=self.grid.shape[1]
        projection_orders=np.concatenate((np.arange(int(n_angular_points/2)+1),-1*np.arange(int(n_angular_points/2)+n_angular_points%2)[:0:-1]))
        
        pos_orders = self.positive_orders
        positive_harmonic_orders=np.array(tuple(self.used_orders.keys()))
        order_ids= np.array(tuple(self.used_orders.values()))
        even_order_mask=positive_harmonic_orders%2==0
        non_zero_mask=positive_harmonic_orders!=0
        order_mask=even_order_mask*non_zero_mask
        harmonic_orders=positive_harmonic_orders[order_mask]
        harmonic_orders_ids=order_ids[order_mask]
        max_order=np.max(harmonic_orders)
            
        SO_order_indices,SO_orders,sorted_order_indices=self.rank_projection_matrix_orders_2d() 
        #sorted_order_indices=self.rank_projection_matrix_orders_2d(projection_vector,self.radial_points[radial_high_pass_index:])
        
        #log.info('sorted_order_indices={}'.format(SO_order_indices))
        first_order_index=SO_order_indices[0]
        first_order=SO_orders[0]#harmonic_orders[first_order_index]
        #log.info('ranked_orders={}'.format(harmonic_orders[sorted_order_indices]))
    
        remaining_rotations=first_order
        current_order=first_order    
        free_orders_mask=True
        angle_coeffs=()
        angles=()
        order_indices=()
        gcds=()
        while remaining_rotations>2:
            #find non invariant orders
            order_multiples=np.arange(current_order,max_order+1,current_order)
            multiple_indices=np.where(np.isin(harmonic_orders,order_multiples))
            free_orders_mask*=~np.isin(sorted_order_indices,multiple_indices)
            #log.info('remaining_orders={}'.format(harmonic_orders[sorted_order_indices[free_orders_mask]]))
            if free_orders_mask.any()==False:
                break
            else:
                #select next order
                current_order_index=sorted_order_indices[free_orders_mask][0]
                current_order=harmonic_orders[current_order_index]
                #log.info('current order ={}'.format(current_order))
                #calculate remaining rotations
                gcd=np.gcd(remaining_rotations,current_order)
                #log.info('gcd={}'.format(gcd))
                #use rotational freedom on current_order
                n_independent_rotations=remaining_rotations/gcd
                #log.info('n independent rotations={}'.format(n_independent_rotations))
                smallest_angle=2*np.pi/n_independent_rotations
                smallest_angle_coeff=np.argmin((np.arange(1,n_independent_rotations)*current_order/gcd)%n_independent_rotations)+1
    
                order_indices+=(current_order_index,)
                angle_coeffs+=(smallest_angle_coeff,)
                angles+=(smallest_angle,)
                gcds+=(gcd,)
                remaining_rotations=gcd
        
        def apply_SO_freedom(harmonic_coefficients,fxs_unknowns):
            #log.info('fxs unknowns shape ={}'.format(fxs_unknowns.shape))
            phases=(-1.j*np.log(fxs_unknowns[order_mask])).real
            #log.info('complete phases ={}'.format(-1.j*np.log(fxs_unknowns[::2])))
            #log.info('phases ={}'.format(phases))
            rotation_phase=0
            for order_index,angle,angle_coeff,gcd in zip(order_indices,angles,angle_coeffs,gcds):
                rotation_phase-=(phases[order_index]//angle)*angle_coeff*angle/gcd
                #log.info('order={} phase={}'.format(harmonic_orders[order_index],phases[order_index]))
                #log.info('min angle*gcd={}'.format(angle))
                #log.info('rotated phase={}'.format((phases[order_index]+harmonic_orders[order_index]*rotation_phase)%(2*np.pi)))
            
            harmonic_coefficients*=np.exp(1.j*projection_orders*rotation_phase)
            return harmonic_coefficients
        return apply_SO_freedom


    # Number of particle estimation experimental not working jet
    def change_n_particles(self,N):
        '''Setter for n_particles attribute'''
        self.n_particles = N

    def get_number_of_particles(self):
        return self.number_of_particles
    
    def apply_particle_number_scaling(self,projection_matrices,zero_id,average_intensity):
        '''
        Let N be the nunmber of particles. this method scales the projection_matrices $V_l$ by $1/\sqrt(N)$ for $l\neq 0$ and $1/N$ for $l = 0$.
        The averaged intensity is scaled by $1/N$, since it is proportional to $V_0$
        '''
        N = settings.project.projections.reciprocal.number_of_particles
        projection_matrices = [matrix/np.sqrt(N) for matrix in projection_matrices]
        projection_matrices[zero_id] = projection_matrices[zero_id]/np.sqrt(N)
        average_intensity = average_intensity/N
        return projection_matrices,average_intensity
            
    def generate_number_of_particles_porjection(self):
        p_opt = settings.project.projections.reciprocal
        zero_id = self.used_orders.get(0,'')
        radial_mask=self.radial_mask[zero_id]
        
        proj_matrices = self.projection_matrices
        if p_opt.use_averaged_intensity:
            I00 = np.abs(self.average_intensity.data.real)
        else:
            I00 = np.abs(proj_matrices[0].flatten().real)
        I00y00=I00/(2*np.sqrt(np.pi))
        N_space = p_opt.number_of_particles.scan_space
        #Ns = np.linspace(*N_space)
        Ns_sqrt = np.linspace(*np.sqrt(N_space[:-1]),N_space[-1])
        Ns = Ns_sqrt**2
        #Ns_lin = np.linspace(*N_space)
        #Ns_lin_sqrt = np.sqrt(Ns_lin)
        #log.info("N_max = {} Ns_sqrt max = {}".format(Ns.max(),Ns_sqrt.max()))
        #log.info("N_min = {} Ns_sqrt min = {}".format(Ns.min(),Ns_sqrt.min()))
        summands = (1/Ns_sqrt-1)[:,None]*I00y00[None,radial_mask]
        #summands = (1/Ns_sqrt-1)[:,None]*I00y00[None,radial_mask]
        nsum=np.sum
        n_pixels=np.prod(self.grid[radial_mask].shape[:-1])
        estimate_n_particles = i_tools.estimate_number_of_particles
        initial_n_particles = p_opt.number_of_particles.initial
        if p_opt.number_of_particles.project:
            def particle_number_projection(I):
                #log.info('I shape = {}'.format(I.shape))
                scaled_I = I[None,radial_mask,...] + summands[:,:,None,None]
                neg_fractions = np.sum(scaled_I<0,axis=(1,2,3))/n_pixels
                grad = np.gradient(neg_fractions,Ns_sqrt)            
                inflection_id = np.argmax(grad)
                self.number_of_particles_dict = {'number_of_particles':Ns[inflection_id],'negative_fraction':neg_fractions,'gradient':grad}
                self.number_of_particles[0] = Ns[inflection_id]
                #log.info('number of particles = {}'.format(self.number_of_particles))
                I[radial_mask] = scaled_I[inflection_id]
                I[I<0]=0
                del(scaled_I)            
                return I
        else:
            def particle_number_projection(I):
                #log.info('I shape = {}'.format(I.shape))
                scaled_I = I[None,radial_mask,...] + summands[:,:,None,None]
                neg_fractions = np.sum(scaled_I<0,axis=(1,2,3))/n_pixels
                grad = np.gradient(neg_fractions,Ns)            
                inflection_id = np.argmax(grad)
                self.number_of_particles_dict = {'number_of_particles':Ns[inflection_id],'negative_fraction':neg_fractions,'gradient':grad}
                self.number_of_particles[0] = Ns[inflection_id]
                #log.info('number of particles = {}'.format(self.number_of_particles))
                #I[radial_mask] += (1/np.sqrt(initial_n_particles)-1)*I00y00[radial_mask]
                #I[I<0]=0
                del(scaled_I)            
                return I
            
        return particle_number_projection
    
    def generate_number_of_particles_porjection_gpu(self):
        p_opt = settings.project.projections.reciprocal
        zero_id = self.used_orders.get(0,'')
        radial_mask=self.radial_mask[zero_id]
        proj_matrices = self.projection_matrices
        if p_opt.use_averaged_intensity:
            I00 = np.abs(self.average_intensity.data.real)
        else:
            I00 = np.abs(proj_matrices[0].flatten().real)
        I00y00=I00/(2*np.sqrt(np.pi))
        N_space = p_opt.number_of_particles.scan_space
        Ns = np.linspace(*N_space)
        summands = (1/np.sqrt(Ns)-1)[:,None]*I00y00[None,radial_mask]
        nsum=np.sum
        n_pixels=np.prod(self.grid[radial_mask].shape)
        estimate_n_particles = i_tools.estimate_number_of_particles


        cld = Multiprocessing.load_openCL_dict()
        cl = cld['cl']
        ctx = cld['context']
        queue = cld['queue']

        kernel = cl.Program(ctx, """
        __kernel void
        count_negative(__global double* I, 
        __global double* summands, 
        __global double* neg_counts, 
        long nN,long nq,long ntheta, long nphi)
        {
        
        long N = get_global_id(0); 
        long q = get_global_id(1); 
        long theta = get_global_id(2); 
        long phi = get_global_id(3); 
        
        // value stores the element that is 
        // computed by the thread
        double neg_count = 0;
        for (int phi = 0; phi < nphi; ++phi)
        {
        double sum = I[q*ntheta*nphi+theta*nphi+phi] + summands[N*nq+q]; 
        neg_count += (1.0-sum/fabs(sum)); // 2 if sum is negative, 0 else
        }
        // Write the matrix to device memory each 
        // thread writes one element
        neg_counts[N*nq*ntheta + q*ntheta + theta] = neg_count/2;
        }    

        __kernel void 
        floatSum(__global float* inVector, __global float* outVector, const int inVectorSize,const long nN, __local float* resultScratch){
        int N = get_global_id(0);
        int gid = get_global_id(1);
        int wid = get_local_id(0);
        int wsize = get_local_size(0);
        int grid = get_group_id(0);
        int grcount = get_num_groups(0);
    
        int i;
        int workAmount = inVectorSize/grcount;
        int startOffest = workAmount * grid + wid;
        int maxOffest = workAmount * (grid + 1);
        if(maxOffset > inVectorSize){
            maxOffset = inVectorSize;
        }
        resultScratch[nN*wsize + wid] = 0.0;
        for(i=startOffest;i<maxOffest;i+=N*wsize){
                resultScratch[N*wsize + wid] += inVector[N*wsize + i];
        }
        barrier(CLK_LOCAL_MEM_FENCE);
    
        if(gid == 0){
        for(i=1;i<wsize;i++){
        resultScratch[0] += resultScratch[i];
        }
        outVector[grid] = resultScratch[0];
        }

        __kernel
        void reduce(__global float* buffer,
        __local float* scratch,
        __const int length,
        __global float* result) {
        int global_index = get_global_id(0);
        float accumulator = INFINITY;
        // Loop sequentially over chunks of input vector
        while (global_index < length) {
        float element = buffer[global_index];
        accumulator = (accumulator < element) ?
        accumulator : element;
        global_index += get_global_size(0);
        }
        // Perform parallel reduction
        int local_id = get_local_id(0)
        scratch[local_id]=accumulator;
        barrier(CLK_LOCAL_MEM_FENCE);
        
        for(int offset = get_local_size(0)/2;
        offset > 0;
        offset = offset / 2) {
        if (local_id<offset){
        double other = scratch[local_id+offset];
        double mine = scratch[local_id];
        scratch[local_id]= (mine<other) ? mine : other;
        }
        barrier(CLK_LOCAL_MEM_FENCE);
        }
        }
    
        sum_space(__global double* neg_count_phi, 
        __global double* neg_count, 
        long nN,long nq,long ntheta, long nphi)
        {
         // Get the index of the current element to be processed
        int N = get_global_id(0);
        int i = get_global_id(1)*2;
        int locali = get_local_id(0);
        int2 va = vload2(i, numbers);
        int2 vb = vload2(i+1, numbers);
        int2 vc = va + vb;
        numReduce[locali] = vc[0] + vc[1];


        long N = get_global_id(0); 
        double sum = 0;
        for (int q = 0; q < nq; ++q)
        {
        for (int theta = 0; theta < ntheta; ++theta)
        {
        sum +=neg_count_phi[N*nq*ntheta + q*ntheta + theta]
        }
        }
        // value stores the element that is 
        // computed by the thread
        double neg_count = 0;
        
        double sum = I[q*ntheta*nphi+theta*nphi+phi] + summands[N*nq+q]; 
        neg_counts[N] += (1.0-sum/fabs(sum)); // 2 if sum is negative, 0 else
        
        // Write the matrix to device memory each 
        // thread writes one element
         = neg_count/2;
        }    
        """).build()
        count_negative = kernel.count_negative
        sum_space = kernel.sum_space
        
        count_negative.set_scalar_arg_dtypes([None,None,None,np.int64,np.int64,np.int64,np.int64])
        
        local_range = None
        nN = len(Ns)
        nq, ntheta, nphi = self.grid[:].shape[:-1]
        nq = np.sum(radial_mask)
        global_range_count = (nN,nq,ntheta)
        global_range_sum = (nN,)
        neg_counts = np.zeros(nN,dtype = float)
        I_mock = np.zeros((nq,ntheta,nphi),dtype = float)

        mf = cl.mem_flags
        summands_buff = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=np.ascontiguousarray(summands))
        I_buff = cl.Buffer(ctx , mf.READ_WRITE, size=I_mock.nbytes)
        out_buff = cl.Buffer(ctx , mf.READ_WRITE, size=neg_counts.nbytes)

        def particle_number_projection(I):
            #log.info('I shape = {} I dtype = {}'.format(I[radial_mask].shape,I.dtype))
            start = time.time()
            cl.enqueue_copy(queue,I_buff,np.ascontiguousarray(I[radial_mask].real))
            count_negative(queue,global_range,local_range,I_buff,summands_buff,out_buff,nN,nq,ntheta,nphi)
            cl.enqueue_copy(queue, out_buff,neg_counts)
            log.info('GPU took {} seconds'.format(time.time()-start))
            grad = np.gradient(neg_counts/n_pixels,Ns)        
            inflection_id = np.argmax(grad)
            self.number_of_particles = Ns[inflection_id]
            #log.info('number of particles = {}'.format(self.number_of_particles))
            I[radial_mask] = I[radial_mask]+summands[inflection_id,:,None,None]
            I[I<0]=0
            return I
        return particle_number_projection




    

# particle number approximation and projection as described in
# K. Pande et.al. PNAS 2018 'Ab initio structure determination from experimental fluctuation X-ray scattering data'
def gnerate_estimate_particle_number(radial_grid,projection_matrices,used_orders):
    '''
    Particle number approximation
    K. Pande et.al. PNAS 2018 'Ab initio structure determination from experimental fluctuation X-ray scattering data'
    :param radial_grid: (N_q) shaped array of radial sampling points.
    :type numpy.ndarray: 
    :param projection_matrices: list of length (L_max+1) containing the $(N_q,min(2*l+1,N_q))$ shaped projection matrices $V_l$
    :type list: of numpy.ndarrays
    :param used_orders: dictionary linking the harmonic orders(as key) to their ids (as value) 
    :type dict:  
    :return N: estimated Number of particles for which the B_l coefficients where calculated.
    :rtype float: 
    '''
    B_l = i_tools.projection_matrices_to_deg2_invariant_3d(projection_matrices)[1:]
    B_l_diag = np.diagonal(B_l,axis1=-2,axis2=-1)
    G = np.sum(B_l_diag**2*(radial_grid**2)[None,:])
    B_l_diag_q2 = np.diagonal(B_l,axis1=-2,axis2=-1)*(radial_grid**2)[None,:]
    calc_Bl=i_tools.spherical_harmonic_coefficients_to_deg2_invariant
    order_ids = list(used_orders.values())
    def estimate_particle_number(I_coeff):
        B_l_guess = calc_Bl(I_coeff)[order_ids][1:]
        B_l_guess_diag = np.diagonal(B_l_guess,axis1=-2,axis2=-1)                
        N = G/np.sum(B_l_guess_diag*B_l_diag_q2)
        log.info('Estimated number of particles = {}'.format(N))
        return 1.0
    return estimate_particle_number

    
    
#positions and point inversion projections
def generate_fix_point_inversion(radial_low_pass = 0.1):
    
    def fix_point_inversion(scattering_amplitude):
        'Routine that corrects point inversion if given scattering amplitudes which only differ by point inversion (i.e. complex conjugation) from each other.'
        phases = scattering_amplitude/np.abs(scattering_amplitude)
        shape = scattering_amplitude.shape
        inversion_indicator = np.sum(phases.imag[:int(shape[0]*radial_low_pass),:shape[1]//2])
        log.info('inversion indicator ={} radial_low_pass = {}'.format(inversion_indicator,radial_low_pass))
        if inversion_indicator<0:
            scattering_amplitude=scattering_amplitude.conjugate()
        return scattering_amplitude
    return fix_point_inversion
def generate_negative_shift_operator(reciprocal_grid,fourier_type):
    dim=reciprocal_grid.n_shape[0]
    pre_exponent=2*np.pi
    pi=np.pi
    if (fourier_type=='Zernike') or (fourier_type=='trapz'):
        pre_exponent=1

    cart_grid = spherical_to_cartesian(reciprocal_grid)
    def negative_shift(reciprocal_density,vector):
        cart_vect = spherical_to_cartesian(vector)
        log.info('cart_vector = {}'.format(cart_vect))
        
        phases = np.exp(-1.j*pre_exponent*(cart_grid*cart_vect).sum(axis=-1))
        log.info('phases shape = {}'.format(phases.shape))
        reciprocal_density*=phases
        return reciprocal_density        
    return negative_shift

def generate_shift_by_operator(grid,opposite_direction=False):
    dim=grid[:].shape[-1]
    pi = np.pi
    if opposite_direction:
        prefactor = -1
    else:
        prefactor = 1
    if dim == 2:
        cart_grid = spherical_to_cartesian(grid)
        def shift_by(reciprocal_density,vector):
            #log.info('vector = {} /n'.format(vector))
            cart_vect = spherical_to_cartesian(vector)
            #log.info('cart shift vect = {}'.format(-cart_vect*prefactor))
            phases = np.exp(-1.j*prefactor*(cart_grid*cart_vect).sum(axis=-1))
            reciprocal_density*=phases
            return reciprocal_density   
    elif dim == 3:
        cart_grid = spherical_to_cartesian(grid)
        def shift_by(reciprocal_density,vector):
            #log.info('vector = {}'.format(vector))
            cart_vect = spherical_to_cartesian(vector)
            #log.info('cart shift vect = {}'.format(-cart_vect*prefactor))
            phases = np.exp(-1.j*prefactor*(cart_grid*cart_vect).sum(axis=-1))
            reciprocal_density*=phases
            return reciprocal_density        
    return shift_by





#from dataclasses import dataclass
import numpy as np
import numpy.ma as mp
import logging

log=logging.getLogger('root')

from xframe.library.gridLibrary import NestedArray
from xframe.library.gridLibrary import uniformGridGetStepSizes
from xframe.library.gridLibrary import SampledFunction
from .misk import returnGrid_deleteFirstDimension



class BCoefficientData(SampledFunction):
    def __init__(self,grid,dataGrid,dimension):
        self.dimension=dimension
        super().__init__(grid,dataGrid)
    def apply(self,function):
        newDataGrid=self.data.apply(function)
        newCrossCorrelationData=BCoefficientData(self.grid,newDataGrid,self.dimension)
        return newCrossCorrelationData
    def apply_along_axis(self,function,axis):
        newDataGrid=self.data.apply_along_axis(function,axis)
#        newGrid=self.grid.swapaxes(axis,len(self.grid.shape)-1)
        newCrossCorrelationData=BCoefficientData(self.grid,newDataGrid,self.dimension)
        return newCrossCorrelationData

class FXS_Data():
    def __init__(self,ccd,incident_wavelength,invariant_extraction_grid,average_intensity=False):
        self.incident_wavelength=incident_wavelength
        self.invariant_extraction_grid=invariant_extraction_grid
        #CCData is supposed to be an instance of SampledFunction
        self.CCD=ccd
        self.aInt=average_intensity
        
        def extractAutocorrelation(self):
            ACD_data=np.moveaxis(self.CCD.diagonal(0,0,1),-1,0).copy()            
            ACD=SampledFunction(invariant_extraction_grid,ACD_data)            
            return ACD

        self.extractAutocorrelation=extractAutocorrelation
        if isinstance(self.CCD,SampledFunction):
            self.ACD=extractAutocorrelation(self)
        else:
            self.ACD=False

        self.dimension=False
        self.bCoeff=False
        self.usedHarmOrders=False
        self.projection_vector=False

'''
@dataclass
class ReciprocalProjectionData:
    average_intensity : object
    proj_matrices : object

@dataclass
class MTIP_Data():
    max_Q : float
    grid : NestedArray = False
    ReciprocalProjectionData
'''

class FTGridPair:
    def __init__(self,realGrid,reciprocalGrid):
        self.realGrid=realGrid
        self.reciprocalGrid=reciprocalGrid



class ProjectionOrders:
    def __init__(orders,ids):
        self.orders=orders
        self.ids=ids
        assert len(orders) == len(ids),'length of orders and ids must match.'

    def __getitem__(self,key):
        return self.orders.__getitem__(key)
    
    def __setitem__(self,key,value):
        self.orders.__setitem__(key,value)



import os
from importlib import util

#install_folder = os.path.abspath(os.path.dirname(__file__)+'/../')+'/'
install_folder = util.find_spec('xframe').submodule_search_locations[0]+'/'
#print(f'xframe is installed at {install_folder}')

default_folder=os.path.expanduser('~/.xframe/')
home_folder=os.path.expanduser('~/.xframe/')
from importlib import util
###Loading User provided config ###
user_config_file = os.path.join(home_folder,'config.py')
if os.path.exists(user_config_file):
    spec = util.spec_from_file_location('file',user_config_file)
    module = util.module_from_spec(spec)
    spec.loader.exec_module(module)
    globals().update(module.__dict__)
home_folder = os.path.expanduser(home_folder)

lazy_denpendency_injection = True

n_control_workers = 0 #8
max_parallel_processes = 200
RAM = 512*1024**3 # in Byte, This value is only used if psutil is not installed
cache_aware = True
L1_cache = 64 #32 # Cache size in kB
L2_cache = 512 #256 # Cache size in kB
loglevel = 'WARNING' #['WARNING','INFO','DEBUG']

default_project_db_name= 'ProjectDB'
default_experiment_db_name= 'ExperimentDB'
default_project_worker_name = 'ProjectWorker'
default_experiment_worker_name = 'ExperimentWorker'
default_experiment_module_name = 'experiment'

default_settings_regexpr = '[-+]?\d*\.*\d+'
worker_regexpr = '^[A-Za-z0-9].*\.py'
settings_version_key = 'settings_version'

load_projects = 'all' # or list of project names
load_experiments = 'all' # or list of experiment names

project_folders=['projects','install_projects']
experiment_folders=['experiments','install_experiments']

log_file = os.path.join(home_folder,'log.txt')
IO = {"folders":
      {"default_home":'~/.xframe/',
       "home":home_folder,
       'install':install_folder,
       'direct':'{path}',
       'experiments':{'home':'experiments/'},
       'projects':
       {'home':'projects/'},
       'default_experiments':
       {'default_home':'experiments/'},
       'default_projects':
       {'default_home':'projects/'},
       'install_experiments':
       {'install':'experiments/'},
       'install_projects':
       {'install':'projects/'},       
       'settings_install_project':
       {'install_projects':'{project}/settings/{worker}/'},
       'settings_install_experiment':
       {'install_experiments':'{experiment}/settings/'},
       'settings_project':
       {'home':'settings/projects/{project}/{worker}/'},
       'settings_experiment':
       {'home':'settings/experiments/{experiment}/'},
       'settings_default_project':
       {'default_home':'settings/projects/{project}/{worker}/'},
       'settings_default_experiment':
       {'default_home':'settings/experiments/{experiment}/'},
       'settings_direct':
           {'direct':'settings/{worker}/'}
       },
      'files':{
          'config':{
              'name':'config.py',
              'folder':'default_home'
          },
          'log':{
              'name':'log.txt',
              'folder':'home'
          },
          'settings_experiment':{
              'name': '{name}.yaml',
              'folder': 'settings_experiment'
          },
          'settings_default_experiment':{
              'name': '{name}.yaml',
              'folder': 'settings_default_experiment'
          },
          'settings_install_experiment':{
              'name': '{name}.yaml',
              'folder': 'settings_install_experiment'
          },
          'settings_project':{
              'name': '{name}.yaml',
              'folder': 'settings_project'
          },
          'settings_default_project':{
              'name': '{name}.yaml',
              'folder': 'settings_default_project'
          },
          'settings_install_project':{
              'name': '{name}.yaml',
              'folder': 'settings_install_project'
          },
          'settings_direct':{
              'name': '{name}.yaml',
              'folder': 'direct'
          }
      }
    }



import click as cl
from xframe.interfaces import ClickInterface

class ClickPlugin(ClickInterface):
    def __init__(self):
        self._click = cl
    @property
    def click(self):
        return self._click
        



import numpy as np
import logging

from xframe.library.gridLibrary import NestedArray
from xframe.library import pythonLibrary as pyLib
from xframe.library.pythonLibrary import get_L2_cache_split_parameters
#from xframe.presenters.matplolibPresenter import heatPolar2D
from xframe.library.mathLibrary import SphericalIntegrator,PolarIntegrator
from xframe import settings
from xframe import Multiprocessing
from .fxs_invariant_tools import harmonic_coeff_to_deg2_invariants_2d
from .fxs_invariant_tools import harmonic_coeff_to_deg2_invariants_3d
from .fxs_invariant_tools import ccd_associated_legendre_matrices_single_m
from .fxs_invariant_tools import ccd_associated_legendre_matrices
from .fxs_invariant_tools import rank_projection_matrices

from xframe.library.pythonLibrary import generate_conjugate
from xframe.library.pythonLibrary import DictNamespace
from xframe.library.physicsLibrary import ewald_sphere_theta_pi
log=logging.getLogger('root')
#pres=heatPolar2D()

### FXS Input Output Methods
class HIOProjection:
    def __init__(self,beta, considered_projections = ['all']):
        self._beta = [beta]
        if not isinstance(considered_projections,(list,tuple)):
            considered_projections = ['all']
        elif len(considered_projections)==0:
            considered_projections = ['all']            
        self.considered_projections = considered_projections
        self.projection = self.generate_HIO()
    @property
    def beta(self):
        return self._beta[0]
    @beta.setter
    def beta(self,value):
        self._beta[0] = value

    def generate_HIO(self):
        beta = self._beta
        where = np.where
        considered_projections = self.considered_projections

        if len(considered_projections)==1:
            name = considered_projections[0]
            def assemble_masks(masks_dict):
                return masks_dict[name]
        else:
            def assemble_masks(masks_dict):
                out_invalid_mask = False
                for name in considered_projections:
                    out_invalid_mask |= mask_dict[name]
                return out_invalid_mask
            
        def hybrid_input_output(without_projection,projection_out,_input):
            out,mask_dict = projection_out
            out_invalid_mask = assemble_masks(mask_dict)
            diff = without_projection-out
            negative_feedback = _input-beta[0]*(diff)
            # negative feedback above is equivalent to Fienup HIO if out is 0 on out_invalid_mask.
            new_input = where(out_invalid_mask,negative_feedback,out)
            return new_input
        return hybrid_input_output

    
def error_reduction(out_without_projection,out,_input):
    return np.array(out[0])


def no_zero_decorator(fun):
    def new_fun(*args,**kwargs):
        val = fun(*args,**kwargs)
        if val <= 0:
            val = 1e-12
        return val
    return new_fun

def get_integrator(dim,grid):
    if dim == 2:
        i = PolarIntegrator(grid[:])
    elif dim == 3:
        i = SphericalIntegrator(grid[:])
    return i

def generate_fxs_error_routine(grid_pair,opt):
    reciprocal_grid = grid_pair.reciprocalGrid
    dim=reciprocal_grid.n_shape[0]
    integrate = get_integrator(dim,reciprocal_grid).integrate_normed
    def error_routine(intensity,projected_intensity):
        diff_norm = integrate(np.square(np.abs(intensity-projected_intensity)))
        proj_norm = integrate(np.square(np.abs(projected_intensity)))
        error = diff_norm/proj_norm
        return error
    return error_routine

def generate_l2_rel_diff_error_routine(grid_pair,_type='real',mask = True):
    if _type == 'reciprocal':
        grid = grid_pair.reciprocalGrid
        pair_id = 0
        power = 1
    else:            
        grid = grid_pair.realGrid
        pair_id = 1
        power = 2
    dim=grid.n_shape[0]
    integrate = get_integrator(dim,grid).integrate

    nabs = np.abs
    nsquare = np.square
    def error_routine(values,projected_values):
        if _type=='real':
            projected_values = projected_values[0]
        #square_diff = (nabs(values - projected_values).real)**2
        #square = (nabs(values).real)**2
        diff = values - projected_values
        square_diff = (diff*diff.conj()).real
        square = (values*values.conj()).real
        square_diff[~mask]=0
        square[~mask]=0
        diff_l2 = integrate(square_diff)
        value_l2 = integrate(square)
        if value_l2!=0:
            error = diff_l2/value_l2
        else:
            error = np.inf
        return error
    return error_routine


def generate_l2_rel_diff_error_routine_cache_aware(grid_pair,L2_cache,_type='real',mask = True):
    if _type == 'reciprocal':
        grid = grid_pair.reciprocalGrid
        pair_id = 0
        power = 1
    else:            
        grid = grid_pair.realGrid
        pair_id = 1
        power = 2
    dim=grid.n_shape[0]
    integrate = get_integrator(dim,grid).integrate

    nabs = np.abs
    nsquare = np.square

    data_shape = grid[:].shape[:-1]
    tmp_square_diff = np.zeros(data_shape,dtype = complex)
    tmp_square = np.zeros(data_shape,dtype = complex)
    #L2 cache divided by two since I will be working on two equally sized matrices tmp_diff and tmp_square_diff
    splitting_index, step = get_L2_cache_split_parameters(data_shape,tmp_square.dtype,L2_cache/2)
    mult = np.multiply

    def loop_0(diff,values):
        for i in range(0,data_shape[0],step):
            i2 = i+step
            d = diff[i:i2]
            v = values[i:i2]
            mult(d,d.conj(),out = tmp_square_diff[i:i2])
            mult(v,v.conj(),out = tmp_square[i:i2])

    def loop_1(diff,values):
        for j in range(data_shape[0]):
            for i in range(0,data_shape[1],step):
                i2 = i+step
                d = diff[j,i:i2]
                v = values[j,i:i2]
                mult(d,d.conj(),out = tmp_square_diff[j,i:i2])
                mult(v,v.conj(),out = tmp_square[j,i:i2])

    def loop_2(diff,values):
        for k in range(data_shape[0]):
            for j in range(data_shape[1]):
                for i in range(0,data_shape[2],step):
                    i2 = i+step
                    d = diff[k,j,i:i2]
                    v = values[k,j,i:i2]
                    mult(d,d.conj(),out = tmp_square_diff[k,j,i:i2])
                    mult(v,v.conj(),out = tmp_square[k,j,i:i2])
                    
    if splitting_index == 0:
        square = loop_0
    elif splitting_index == 1:
        square = loop_1
    elif splitting_index == 2:
        square = loop_2
        
    def error_routine(values,projected_values):
        if _type=='real':
            projected_values = projected_values[0]
        diff = values - projected_values
        square(diff,values)
        tmp_square_diff[~mask]=0
        tmp_square[~mask]=0
        diff_l2 = integrate(tmp_square_diff.real)
        #diff_l2 = np.sum(tmp_square_diff.real)
        value_l2 = integrate(tmp_square.real)
        #value_l2 = np.sum(tmp_square.real)
        if value_l2!=0:
            error = diff_l2/value_l2
        else:
            error = np.inf
        return error
    
    if splitting_index<0:
        error_routine = generate_l2_rel_diff_error_routine(grid_pair,_type=_type)
    return error_routine


def generate_l2_rel_diff_error_routine_gpu(grid_pair,_type='real'):
    if _type == 'reciprocal':
        grid = grid_pair.reciprocalGrid
        pair_id = 0
        power = 1
    else:            
        grid = grid_pair.realGrid
        pair_id = 1
        power = 2
    dim=grid.n_shape[0]
    integrate = get_integrator(dim,grid).integrate

    nabs = np.abs
    nsquare = np.square
    shape = grid[:].shape[:-1]
    size = np.prod(shape)
    
    kernel_str = """
    __kernel void
    calc_squares(
     double2* in_val, 
     double2* in_projected_val, 
     double* out_val_abs,
     double* out_diff_abs,
     long nr_ntheta,
     long ntheta
    )
    {
    
    long i = get_global_id(0); 
    long j = get_global_id(1);
    long k = get_global_id(2); 
    long index = i*nr_ntheta+j*ntheta+k;

    //compute abs^2 of in_val
    double2 val = in_val[index];
    out_val_abs[index] = pown(val.x,2)+pown(val.y,2);
    //compute abs^2 of in_projected_val-in_val
    double2 diff = in_projected_val[index]-val;
    out_diff_abs[index] =  pown(diff.x,2)+pown(diff.y,2);   

    }  
    """

    kernel_dict={
            'kernel': kernel_str,
            'name': 'error_squares',
            'functions': [{
                'name': 'calc_squares',
                'dtypes' : [complex,complex,float,float,np.int64,np.int64],
                'shapes' : [shape,shape,shape,shape,None,None],
                'arg_roles' : ['input','input','output','output','const_input','const_input'],
                'const_inputs' : [None,None,None,None,shape[0]*shape[1],shape[1]],
                'global_range' : shape,
                'local_range' : None
            }]
        }

    calc_error_squares_process =  Multiprocessing.openCL_plugin.ClProcess(kernel_dict)
    calc_error_squares = Multiprocessing.comm_module.add_gpu_process(calc_error_squares_process)
    
    def error_routine(values,projected_values):
        #square_diff = (nabs(values - projected_values).real)**2
        #square = (nabs(values).real)**2
        square,square_diff =  calc_error_squares(values,projected_values)
        #log.info(square)
        diff_l2 = integrate(square_diff)
        value_l2 = integrate(square)
        if value_l2!=0:
            error = diff_l2/value_l2
        else:
            error = np.inf
        #log.info('error = {}'.format(error))
        return error
    log.info('created error routine.')
    return error_routine


def generate_real_l2_rel_diff_error_routine(grid_pair,**kwargs):
    use_gpu = settings.project.GPU.get('error_squares',False)
    limit_to_initial_mask = kwargs.get('inside_initial_support',False)
    if limit_to_initial_mask:
        #log.info('limit to initial support \n\n\n\n') 
        mask = kwargs['initial_mask']
    else:
        mask = True
        
    if settings.general.cache_aware:
        error_routine = generate_l2_rel_diff_error_routine_cache_aware(grid_pair,settings.general.L2_cache,_type='real',mask = mask)
    else:
        error_routine = generate_l2_rel_diff_error_routine(grid_pair,_type='real',mask = mask)
    return error_routine
def generate_reciprocal_l2_rel_diff_error_routine(grid_pair,**kwargs):
    use_gpu = settings.project.GPU.get('error_squares',False)
    if settings.general.cache_aware:
        error_routine = generate_l2_rel_diff_error_routine_cache_aware(grid_pair,settings.general.L2_cache,_type='reziprocal')
    else:
        error_routine = generate_l2_rel_diff_error_routine(grid_pair,_type='reciprocal')
    def error_correct_arguments(values,projected_values,harmonic_coefficients):
        return error_routine(values,projected_values)
    return error_correct_arguments


def generate_deg2_invariant_l2_diff(grid_pair,**opt):
    reference_invariant = opt['deg2_invariants']
    #log.info('Bm shape = {}'.format(reference_invariant.shape))
    used_orders = opt['used_orders']
    #log.info('\n\n used_orders = {}'.format(used_orders))
    #log.info('deg 2 invariant shape = {}\n'.format(reference_invariant.shape))
    n_particles = opt['n_particles']
    invariant_mask = opt['invariant_mask']
    #log.info('all qs valid = {}'.format(invariant_mask.all()))
    reference_invariant=reference_invariant[np.array(list(used_orders.keys())).astype(int)]
    grid = grid_pair.reciprocalGrid[:]
    dimensions = grid.shape[-1]
    radial_points = grid.__getitem__((slice(None),)+(0,)*dimensions) # expression is just grid[:,0,0,0] in 3D case ans grid[:,0,0] in 2D
    if dimensions == 2:
        invariant_error = _generate_deg2_invariant_diff_2d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask)
    elif dimensions == 3:
        invariant_error = _generate_deg2_invariant_diff_3d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask)
    return invariant_error
def generate_deg2_ranked_invariant_l2_diff(grid_pair,**opt):
    reference_invariant = opt['deg2_invariants']
    projection_matrices = opt['projection_matrices']
    #log.info('Bm shape = {}'.format(reference_invariant.shape))
    used_orders = opt['used_orders']
    order_array=np.array(list(used_orders.keys())).astype(int)    
    #log.info('\n\n used_orders = {}'.format(used_orders))
    #log.info('deg 2 invariant shape = {}\n'.format(reference_invariant.shape))
    n_particles = opt['n_particles']
    invariant_mask = opt['invariant_mask']
    #log.info('all qs valid = {}'.format(invariant_mask.all()))
    reference_invariant=reference_invariant[np.array(list(used_orders.keys())).astype(int)]
    grid = grid_pair.reciprocalGrid[:]
    dimensions = grid.shape[-1]
    radial_points = grid.__getitem__((slice(None),)+(0,)*dimensions) # expression is just grid[:,0,0,0] in 3D case ans grid[:,0,0] in 2D
    error_order = opt.get('order',False)
    no_error_order_specified = (not isinstance(error_order,int)) or isinstance(error_order,bool)
    
    if dimensions == 2:        
        invariant_error = _generate_deg2_invariant_diff_2d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask)
        if no_error_order_specified:
            ranked_order_indices,ranked_orders,metric = rank_projection_matrices(2,projection_matrices,order_array,radial_points)
            error_id = ranked_order_indices[0]
        else:
            error_id = used_orders[error_order]            
                                                                             
    elif dimensions == 3:
        invariant_error = _generate_deg2_invariant_diff_3d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask)
        error_order = opt.get('order',False)
        if no_error_order_specified:
            ranked_order_indices,ranked_orders,metric = rank_projection_matrices(3,projection_matrices,order_array,radial_points)
            error_id = ranked_order_indices[0]
        else:
            error_id = used_orders[error_order]
            
    def ranked_deg2_error(values,projected_values,Ilms):
        invariant_errors = invariant_error(values,projected_values,Ilms)
        #log.info('error_order = {}'.format(error_id))
        return invariant_errors[error_id]
    return ranked_deg2_error
def _generate_deg2_invariant_diff_2d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask):
    qq = radial_points[None,:,None]*radial_points[None,None,:]
    reference_masked = reference_invariant.copy()
    reference=reference_masked.copy()
    #log.info(qq)
    norm= np.sum(reference*reference.conj(),axis=(1,2)).real
    #log.info(norm)
    non_zero_mask = norm !=0
    n_orders=len(norm)
    errors = np.full(n_orders,-1,dtype = float)
    non_zero_norm = norm[non_zero_mask]
    order_array = np.array(tuple(used_orders.values()))    
    mask = np.zeros(reference_masked.shape,dtype = bool)
    zero_id = used_orders[0]
    def invariant_error(values,projected_values,Ims):
        Bm =  harmonic_coeff_to_deg2_invariants_2d(Ims)
        #log.info('Bl0-ref0 = {}'.format(reference_masked/sqrt(n_particles[0])))
        #log.info("bm shape/type = {}/{} reference shape/type = {}/{}".format(Bm.shape,Bm.dtype,reference.shape,reference.dtype))
        Bm= Bm[order_array].copy()
        #Bm[mask] = 0
        #log.info('ref0 = {}'.format(reference_masked[0,1,:10]/n_particles[0]))
        #log.info('values= {}'.format(Ims[9][1,:10]))
        #log.info('Bl0-ref0= {}'.format((Bl[0,1,:10]*n_particles[0]-reference_masked[0,1,:10])/reference_masked[0,1,:10]))
        reference[zero_id] = reference_masked[zero_id]/n_particles[0]   
        diff = reference - Bm
        #squared_scaled_diff = (qq*diff*diff.conj()).real
        squared_scaled_diff = (diff*diff.conj()).real
        norm_diff = np.sum(squared_scaled_diff,axis=(1,2))
        errors[non_zero_mask] = norm_diff[non_zero_mask]/non_zero_norm
        return errors.copy()
    
    def invariant_error_old(values,projected_values,Ims):
        Bm = harmonic_coeff_to_deg2_invariants_2d(Ims)
        diff = reference_invariant - Bm[order_array]
        errors[non_zero_mask] = np.sum(qq*diff*diff.conj(),axis=(1,2))[non_zero_mask]/non_zero_norm
        return errors.copy()
    return invariant_error

def _generate_deg2_invariant_diff_3d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask):
    qq = (radial_points[None,:,None]*radial_points[None,None,:])**2
    #reference_invariant=reference_invariant[:,30:,30:]
    #log.info('deg2 shape = {}'.format(reference_invariant.shape))
    #log.info('first values = {}'.format(reference_invariant[0,1,:10]))
    order_array = np.array(tuple(used_orders.values()))
    log.info('used orders = {}'.format(order_array))
    zero_id = used_orders[0]
    reference_masked = reference_invariant.copy()
    #log.info('refference shape = {}'.format(reference_masked.shape))
    mask = np.zeros(reference_masked.shape,dtype = bool)
    #log.info('invariant mask shape = {}, containes unmasked points = {}'.format(invariant_mask.shape,invariant_mask.any()))
    mask[:] = ~invariant_mask[order_array]
    #log.info('mask shape = {} '.format(reference_masked.shape))
    reference_masked[mask]=0
    reference=reference_masked.copy()
    sqrt = np.sqrt
    #norm= np.sum(qq*reference*reference.conj(),axis=(1,2))
    norm= np.sum(reference*reference.conj(),axis=(1,2))
    non_zero_mask = norm !=0
    n_orders=len(norm)
    errors = np.full(n_orders,-1,dtype = float)
    non_zero_norm = norm[non_zero_mask]
    log.info('first values = {}'.format(reference[0,1,:10]))
    def invariant_error(values,projected_values,Ims):
        Bl =  harmonic_coeff_to_deg2_invariants_3d(Ims)
        #log.info('Bl0-ref0 = {}'.format(reference_masked/sqrt(n_particles[0])))
        Bl= Bl[order_array].copy()
        Bl[mask] = 0
        #log.info('ref0 = {}'.format(reference_masked[0,1,:10]/n_particles[0]))
        #log.info('values= {}'.format(Ims[9][1,:10]))
        #log.info('Bl0-ref0= {}'.format((Bl[0,1,:10]*n_particles[0]-reference_masked[0,1,:10])/reference_masked[0,1,:10]))
        reference[zero_id] = reference_masked[zero_id]/n_particles[0]   
        diff = reference - Bl
        #squared_scaled_diff = (qq*diff*diff.conj()).real
        squared_scaled_diff = (diff*diff.conj()).real
        norm_diff = np.sum(squared_scaled_diff,axis=(1,2))
        errors[non_zero_mask] = norm_diff[non_zero_mask]/non_zero_norm
        return errors.copy()
    return invariant_error

def _generate_deg2_invariant_diff_3d_old(radial_points,reference_invariant,used_orders):
    qq = (radial_points[None,:,None]*radial_points[None,None,:])**2
    #reference_invariant=reference_invariant[:,30:,30:]
    norm= np.sum(reference_invariant*reference_invariant.conj(),axis=(1,2))
    non_zero_mask = norm !=0
    n_orders=len(norm)
    errors = np.full(n_orders,-1,dtype = float)
    non_zero_norm = norm[non_zero_mask]
    order_array = np.array(tuple(used_orders.values()))
    def invariant_error(Ims):
        Bl = harmonic_coeff_to_deg2_invariants_3d(Ims)
        #Bl = Bl[:,30:,30:]
        #log.info('len ref = {} len bl = {}'.format(len(reference_invariant),len(Bl)))
        diff = reference_invariant - Bl[order_array]
        norm_diff = np.sum(diff*diff.conj(),axis=(1,2))
        #log.info(norm_diff)
        errors[non_zero_mask] = norm_diff[non_zero_mask]/non_zero_norm
        #log.info(errors)
        return errors.copy()
    return invariant_error



def generate_fqc_error(grid_pair,**opt):
    reference_invariant = opt['deg2_invariants']
    used_orders = opt['used_orders']
    #log.info('\n\n used_orders = {}'.format(used_orders))
    #log.info('deg 2 invariant shape = {}\n'.format(reference_invariant.shape))
    n_particles = opt['n_particles']
    invariant_mask = opt['invariant_mask']
    xray_wavelength = opt['xray_wavelength']
    log.info('all qs valid = {}'.format(invariant_mask.all()))
    reference_invariant=reference_invariant[np.array(list(used_orders.keys())).astype(int)]
    grid = grid_pair.reciprocalGrid[:]
    dimensions = grid.shape[-1]
    radial_points = grid.__getitem__((slice(None),)+(0,)*dimensions) # expression is just grid[:,0,0,0] in 3D case ans grid[:,0,0] in 2D
    if dimensions == 2:
        invariant_error = _generate_fqc_2d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask,xray_wavelength)
    elif dimensions == 3:
        invariant_error = _generate_fqc_3d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask,xray_wavelength)
    return invariant_error
def _generate_fqc_2d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask,xray_wavelength):
    qq = radial_points[None,:,None]*radial_points[None,None,:]
    log.info(qq)
    norm= np.sum(qq*reference_invariant*reference_invariant.conj(),axis=(1,2))
    log.info(norm)
    non_zero_mask = norm !=0
    n_orders=len(norm)
    errors = np.full(n_orders,-1,dtype = float)
    non_zero_norm = norm[non_zero_mask]
    order_array = np.array(tuple(used_orders.values()))
    def invariant_error(values,projected_values,Ims):
        Bm = harmonic_coeff_to_deg2_invariants_2d(Ims)
        diff = reference_invariant - Bl[order_array]
        errors[non_zero_mask] = np.sum(qq*diff*diff.conj(),axis=(1,2))[non_zero_mask]/non_zero_norm
        return errors.copy()
    return invariant_error

def _generate_fqc_3d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask,xray_wavelength):
    order_array = np.array(tuple(used_orders.keys())).astype(int)
    order_ids = np.array(tuple(used_orders.values()))
    max_order = np.max(order_array)
    thetas = ewald_sphere_theta_pi(xray_wavelength,radial_points)    
    P_matrix = np.moveaxis(ccd_associated_legendre_matrices(thetas,max_order,max_order),-1,0)

    reference_bl_masked = reference_invariant.copy()
    mask = np.zeros(reference_bl_masked.shape,dtype = bool)
    mask[:] = ~invariant_mask[order_ids]
    reference_bl_masked[mask]=0
    
    def calc_ccn(bl):
        return np.sum(bl[1:,...,None]*P_matrix[1:],axis = 0)
    def calc_2ccn_average(c1,c2):
        return (c1[...,0]*c2[...,0]).real + 2*np.sum(c1[...,1:]*c2[...,1:].conj(),axis = -1).real
    
    reference_ccn = calc_ccn(reference_bl_masked)
    reference_average = calc_2ccn_average(reference_ccn,reference_ccn)
    reference_weights = (P_matrix[1:,...,0]*reference_ccn[None,...,0]).real + 2*np.sum(P_matrix[1:,...,1:]*reference_ccn[None,...,1:].conj(),axis = -1).real    
    
    sqrt = np.sqrt
    fqc = np.ones_like(reference_average)
    
    def fqc_error(values,projected_values,Ims):
        Bl =  harmonic_coeff_to_deg2_invariants_3d(Ims)
        #Bl = np.random.rand(*(Bl_r.shape)).astype(Bl_r.dtype)
        #Bl[0]=Bl_r[0]
        Bl[mask]=0
        ccn = calc_ccn(Bl)
        average = calc_2ccn_average(ccn,ccn)
        norm = sqrt(average*reference_average)
        positive_mask = (norm>=0)

        #control_average = calc_2ccn_average(ccn,reference_ccn)
        control_average = np.sum(Bl[1:]*reference_weights,axis = 0)
        
        fqc[positive_mask] = control_average[positive_mask]/norm[positive_mask]
        fqc[~positive_mask] = 1
        errors = tuple(1-np.mean(fqc[i,:i+1]) for i in range(len(radial_points)))        
        return np.array(errors)
    return fqc_error



def generate_II_error(grid_pair,**opt):
    reference_invariant = opt['deg2_invariants']
    used_orders = opt['used_orders']
    #log.info('\n\n used_orders = {}'.format(used_orders))
    #log.info('deg 2 invariant shape = {}\n'.format(reference_invariant.shape))
    n_particles = opt['n_particles']
    invariant_mask = opt['invariant_mask']
    xray_wavelength = opt['xray_wavelength']
    log.info('all qs valid = {}'.format(invariant_mask.all()))
    reference_invariant=reference_invariant[np.array(list(used_orders.keys())).astype(int)]
    grid = grid_pair.reciprocalGrid[:]
    dimensions = grid.shape[-1]
    radial_points = grid.__getitem__((slice(None),)+(0,)*dimensions) # expression is just grid[:,0,0,0] in 3D case ans grid[:,0,0] in 2D
    if dimensions == 2:
        invariant_error = _generate_II_2d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask,xray_wavelength)
    elif dimensions == 3:
        invariant_error = _generate_II_3d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask,xray_wavelength)
    return invariant_error
def _generate_II_2d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask,xray_wavelength):
    qq = radial_points[None,:,None]*radial_points[None,None,:]
    log.info(qq)
    norm= np.sum(qq*reference_invariant*reference_invariant.conj(),axis=(1,2))
    log.info(norm)
    non_zero_mask = norm !=0
    n_orders=len(norm)
    errors = np.full(n_orders,-1,dtype = float)
    non_zero_norm = norm[non_zero_mask]
    order_array = np.array(tuple(used_orders.values()))
    def invariant_error(values,projected_values,Ims):
        Bm = harmonic_coeff_to_deg2_invariants_2d(Ims)
        diff = reference_invariant - Bl[order_array]
        errors[non_zero_mask] = np.sum(qq*diff*diff.conj(),axis=(1,2))[non_zero_mask]/non_zero_norm
        return errors.copy()
    return invariant_error

def _generate_II_3d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask,xray_wavelength):
    order_array = np.array(tuple(used_orders.keys())).astype(int)
    order_ids = np.array(tuple(used_orders.values()))
    max_order = np.max(order_array)
    thetas = ewald_sphere_theta_pi(xray_wavelength,radial_points)    
    P_matrix = np.moveaxis(ccd_associated_legendre_matrices(thetas,max_order,max_order),-1,0)

    reference_bl_masked = reference_invariant.copy()
    mask = np.zeros(reference_bl_masked.shape,dtype = bool)
    mask[:] = ~invariant_mask[order_ids]
    reference_bl_masked[mask]=0

    # The used spherical harmonic transform already include the factor sqrt(4 \pi/(2l+1)) in the spherical harmonics they are orthonormal.
    # This factor is as well included in the extraction routine for the Bl coefficients, which means we only need to sum over the Bl coefficients.
    reference_II = np.sum(reference_bl_masked[1:],axis = 0)
    reference_II_squared = (reference_II*reference_II.conj()).real
    positive_mask = reference_II_squared>0
    current_II = np.zeros_like(reference_II)
    #errors = np.zeros_like(reference_II)
    qq = (radial_points[:,None]*radial_points[None,:])**2
    
    def II_error(values,projected_values,Ims):
        #log.info('fuuuuuuu')
        #print("fu")
        Bl =  harmonic_coeff_to_deg2_invariants_3d(Ims)
        Bl[mask]=0
        np.sum(Bl[1:],axis = 0,out = current_II)
        diff = reference_II-current_II
        square_diff = (diff*diff.conj()).real
        #log.info('considered fraction {}'.format(np.sum(positive_mask)/np.prod(positive_mask.shape)))

        #errors=square_diff[positive_mask]/reference_II_squared[positive_mask]
        #error = np.median(errors)
        
        #error = np.sum(square_diff[positive_mask]*qq[positive_mask])/np.sum(reference_II_squared[positive_mask]*qq[positive_mask])

        error = 1- np.sum(current_II*reference_II*qq)/np.sqrt(np.sum(current_II**2*qq)*np.sum(reference_II**2*qq))
        return error
    return II_error


def generate_ccd_diff(grid_pair,**opt):
    '''
    calculates the difference in l_2 norm between the angularly integrated averaged cross-correlation $\frac{1}{2\pi} \int_0^{2\pi} C(q,q',\phi) d\phi$( or equivalently the zeros harmonic coefficient $C_0(q,q')$ of the averaged cross-correlation ) calculated from reconstructed intensity and the one coming from the input parameters.
    All this is possible due to the following relation $$ C_0(q,q') =  \sum_l B_l \overline{P}^0_l(q) \overline{P}^0_l(q') $$.
    Number of particles scaling is taken into account.
    min_order: Defines the starting order in the sum $\sum_l B_l \overline{P}^0_l(q) \overline{P}^0_l(q')$. Can be used to exclude 0'th order.
    '''
    reference_invariant = opt['deg2_invariants']
    C_order = opt['C_order']
    used_orders = opt['used_orders']
    n_particles = opt['n_particles']
    invariant_mask = opt['invariant_mask']
    xray_wavelength = opt['xray_wavelength']
    reference_invariant=reference_invariant[np.array(list(used_orders.keys())).astype(int)]
    grid = grid_pair.reciprocalGrid[:]
    dimensions = grid.shape[-1]
    radial_points = grid.__getitem__((slice(None),)+(0,)*dimensions) # expression is just grid[:,0,0,0] in 3D case ans grid[:,0,0] in 2D
    if dimensions == 2:
        NotImplementedError('ccd error routine is not jet implemented for 2D reconstructions.')
    elif dimensions == 3:
        invariant_error = _generate_ccd_diff_3d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask,C_order,xray_wavelength)
    return invariant_error

def _generate_ccd_diff_3d(radial_points,reference_invariant,used_orders,n_particles,invariant_mask,C_order,xray_wavelength):
    #qq = (radial_points[None,:,None]*radial_points[None,None,:])**2        
    order_array = np.array(tuple(used_orders.values()))
    zero_id = used_orders[0]

    relevant_orders_mask = order_array>=C_order
    thetas = ewald_sphere_theta_pi(xray_wavelength,radial_points)                       
    PlmPlm = np.moveaxis(np.squeeze(ccd_associated_legendre_matrices_single_m(thetas,np.max(order_array),C_order)),-1,0)[order_array]
    PlmPlm[zero_id] = 0
    PlmPlm = PlmPlm[relevant_orders_mask]
    PlmPlm[np.isnan(PlmPlm)] =0
    
    reference_masked = reference_invariant.copy()    
    mask = np.zeros(reference_masked.shape,dtype = bool)    
    mask[:] = ~invariant_mask[order_array]    
    reference_masked[mask]=0    
    reference=reference_masked.copy()
    ref_C_m = np.sum(reference_masked[relevant_orders_mask]*PlmPlm,axis = 0)
    norm= np.sum(ref_C_m*ref_C_m.conj())
    assert norm !=0,'C_m seems to be zero, abbort creation of error routine.'
    
    sqrt = np.sqrt    
    def ccd_error(values,projected_values,Ims):
        Bl =  harmonic_coeff_to_deg2_invariants_3d(Ims)[order_array]
        Bl[mask] = 0
        Bl[zero_id] *= sqrt(n_particles[0])
        C_m_diff = np.sum(Bl[relevant_orders_mask]*PlmPlm,axis = 0)-ref_C_m
        squared_scaled_diff = (C_m_diff*C_m_diff.conj()).real
        norm_diff = np.sum(squared_scaled_diff)
        error = norm_diff/norm
        return error
    return ccd_error


def generate_support_size(grid_pair,**opt):    
    def support_size(values,projected_values):
        return np.sum((values.real == projected_values.real)**2)/np.prod(values.shape)
    return support_size
    
error_generators={'real':
                  {'l2_projection_diff':generate_real_l2_rel_diff_error_routine,
                   'support_size':generate_support_size
                   },
                  'reciprocal':{                  
                      'l2_projection_diff':generate_reciprocal_l2_rel_diff_error_routine,
                      'deg2_invariant_l2_diff':generate_deg2_invariant_l2_diff,
                      'ccd_diff':generate_ccd_diff,
                      'fqc_error':generate_fqc_error,
                      'II_error':generate_II_error,
                      'deg2_ranked_invariant_l2_diff':generate_deg2_ranked_invariant_l2_diff
                  }
                  }
error_number_of_arguments = {'real':2,'reciprocal':3}

def generate_error_routines(error_opt,grid_pair,**kwargs):
    routine_names = error_opt.methods    
    #log.info('error opts = {}'.format(routine_options))
    error_routines={}
    #log.info('gen_errors {}'.format(routine_names))
    for category,copt in routine_names.items():
        if not 'calculate' in copt:
            continue
        name_list = copt.calculate
        if isinstance(name_list,list):
            error_routines[category]={}            
            for err_name in name_list:
                generator = error_generators[category][err_name]
                opt = copt.get(err_name,DictNamespace()).dict()
                #log.info(opt)
                kwargs.update(opt)
                error_routines[category][err_name] = generator(grid_pair,**kwargs)
    category_routines={}
    for category,routines in error_routines.items():
        if len(routines)>=1:
            combined_routine = combine_error_routines(routines)
        else:
            def mock_routine(*args):
                return {}
            combined_routine = mock_routine
        category_routines[category] = [combined_routine,error_number_of_arguments[category]]

    #log.info('return all error routines')
    return category_routines

def combine_error_routines(error_function_dict,n_arguments=2):
    ''' Returns callable that calculates all error routines in error_function_dict and stores their values in a dict with the same keys.
    If error_flnction_dict containes no entries it just returns False'''
    def error_metrics(*args):
        #log.info('len error arguments = {}'.format(len(args)))
        errors={}
        for key,f in error_function_dict.items():
            errors[key]=f(*args)
        return errors
    return error_metrics

def generate_main_error_routine(error_names,_type):    
    if _type == 'mean':
        method = np.mean
    elif _type == 'min':
        method = np.min
    elif _type == 'max':
        method = np.max
    elif _type == 'prod':
        method = np.prod
    
    def main_error(error_dict):
        try:
            real_errors = [error_dict['real'][name][-1] for name in error_names['real']]
            reciprocal_errors = [error_dict['reciprocal'][name][-1] for name in error_names['reciprocal']]
            err_array=np.array(real_errors+reciprocal_errors)
            return method(err_array)
        except IndexError:
            return -1
        #return method(err_array)
    return main_error



import numpy as np
from scipy.special import roots_legendre
import logging

log=logging.getLogger('root')

from xframe.library import mathLibrary as mLib
from xframe.library.mathLibrary import  polar_spherical_dft_reciprocity_relation_radial_cutoffs
from xframe.library import pythonLibrary as pyLib
from xframe.library.pythonLibrary import DictNamespace
from xframe.library.gridLibrary import NestedArray
from xframe.library.gridLibrary import GridFactory
from xframe.library.gridLibrary import uniformGrid_func
from .classes import FTGridPair
from xframe import settings
from .misk import _get_reciprocity_coefficient



def max_order_from_n_angular_steps(dimensions,n_phi,anti_aliazing_degree=2):
    '''
    max_order for highest power of 2 smaller than n_phi
    '''
    n_phi = 2**int(np.log2(n_phi))
    if dimensions == 2:
        max_order = (n_phi -1)//2
    elif dimensions == 3:
        N = anti_aliazing_degree
        max_order = n_phi//(N+1)
    return max_order

def n_angular_step_from_max_order(dimensions,max_order,anti_aliazing_degree=2):
    dim = dimensions
    max_order = max_order
    size_dict={}
    if dim == 2:
        n_phi = 2**(int(np.log2(max_order*2+1))+1)
        size_dict['n_phi'] = n_phi
    elif dim == 3 :
        N = anti_aliazing_degree
        n_phi = 2**(int(np.log2((N+1)*max_order)) + 1)
        n_theta = n_phi//2
        size_dict['n_phi'] = n_phi
        size_dict['n_phi'] = n_theta


### FT grid Pairs
def polarFFTGridPair_Bessel_uniform(realCutOff,N1,N2):
    def reciprocalGrid():
        besselZeros=np.arange(1,N1,1)
        halfBesselOrders=np.arange(np.floor(N2/2)+1)
        besselOrders=np.concatenate((halfBesselOrders[:-1],-1*halfBesselOrders[1:]))
  #      log.info('besselOrders shape = {}'.format(besselOrders))                
        radialPositions=mLib.besselZeros(besselOrders,besselZeros)
        
        radialPositions=radialPositions.flatten()/realCutOff
#        log.info('radial Positions shape = {}'.format(radialPositions.shape))
        polarNodes=[2*np.pi*besselOrders/N2]
        repeatedPolarNodes=np.repeat(polarNodes,N1-1,axis=0).flatten()
#        log.info('polar nodes shape ={}'.format(repeatedPolarNodes.shape))
        gridArray=np.stack((radialPositions,repeatedPolarNodes),axis=1).reshape(N1-1,N2,2)
        gridArray=pyLib.getArrayOfArray(gridArray)
        gridShape=[np.array([N1-1]),np.full(N1-1,N2)]
        grid=NestedArray(gridArray,1)
        return grid

    def realGrid():
        besselZeros=np.arange(1,N1+1,1)
        halfBesselOrders=np.arange(np.floor(N2/2)+1)
        besselOrders=np.concatenate((halfBesselOrders[:-1],-1*halfBesselOrders[1:]))
#        log.info('besselOrders = {}'.format(besselOrders))
        zeros=mLib.besselZeros(besselOrders,besselZeros)
#        log.info('zeros={}'.format(zeros))
        zeros.shape=(N2,N1)
        zeros=zeros.T
#        log.info('zeros shaped ={}'.format(zeros))
        #       log.info('zeros={}'.format(zeros))
        denominators=zeros[:-1,:].flatten()*realCutOff
#        log.info('denominators={}'.format(denominators))
        divisors=np.repeat([zeros[-1,:].flatten()],N1-1,axis=0).flatten()
#        log.info('divisors={}'.format(divisors))
        radialPositions=denominators/divisors
#        log.info('radial Positions ={}'.format(radialPositions))
        radialPositions=radialPositions.flatten()
        polarNodes=[2*np.pi*besselOrders/N2]
        repeatedPolarNodes=np.repeat(polarNodes,N1-1,axis=0).flatten()

        gridArray=np.stack((radialPositions,repeatedPolarNodes),axis=1).reshape(N1-1,N2,2)
        gridArray=pyLib.getArrayOfArray(gridArray)
        gridShape=[np.array([N1-1]),np.full(N1-1,N2)]        
        grid=NestedArray(gridArray,1)
#        log.info('grid array={}'.format(grid.array))
        return grid

    ftGridPair=FTGridPair(realGrid(),reciprocalGrid())
    return ftGridPair

def polar_bessel(realCutOff,N1,N2):
    bessel_zeros=np.arange(1,N1+1)
    max_order=int((N2-N2%2)/2)
    orders=np.concatenate((np.arange(max_order+1),-1*np.arange(max_order+N2%2)[:0:-1]))
    zeros=mLib.besselZeros(orders,bessel_zeros)    
#    zeros[1:]=zeros[1:]/(np.abs(orders[1:,None])*np.pi/2)-0.86 #min for 50:
#    zeros[1:]=zeros[1:]/(np.abs(orders[1:,None])*np.pi/2)-0.72 #smallest center
#    zeros[1:]=zeros[1:]-np.abs(orders[1:,None])*np.pi/2*0.95 # q min for 50
    def phi_func():
        return orders*np.pi/max_order
    
    def reciprocalGrid():
        def q_func(*args):
            return (zeros)/realCutOff
            
        reciprocal_grid=GridFactory.construct_grid('uniform_dependent',(phi_func,q_func))
        reciprocal_grid.array=np.swapaxes(reciprocal_grid.array,0,1)[...,::-1]
        return reciprocal_grid

    def realGrid():
        def r_func(*args):
            #return (zeros*realCutOff)/zeros[:,-1,None]
            return (zeros)
        real_grid=GridFactory.construct_grid('uniform_dependent',(phi_func,r_func))
        real_grid.array=np.swapaxes(real_grid.array,0,1)[...,::-1]
        return real_grid

    ftGridPair=FTGridPair(realGrid(),reciprocalGrid())
    return ftGridPair

def polar_bessel2(realCutOff,N1,N2):
    bessel_zeros=np.arange(1,N1+1)
    max_order=int((N2-N2%2)/2)
    orders=np.concatenate((np.arange(max_order+1),-1*np.arange(max_order+N2%2)[:0:-1]))
    zeros=mLib.besselZeros(orders,bessel_zeros)    
#    zeros[1:]=zeros[1:]/(np.abs(orders[1:,None])*np.pi/2)-0.86 #min for 50:
#    zeros[1:]=zeros[1:]/(np.abs(orders[1:,None])*np.pi/2)-0.72 #smallest center
#    zeros[1:]=zeros[1:]-np.abs(orders[1:,None])*np.pi/2*0.95 # q min for 50
    def phi_func():
        return orders*np.pi/max_order
    
    def reciprocalGrid():
        def q_func(*args):
            return (zeros)/realCutOff
            
        reciprocal_grid=GridFactory.construct_grid('uniform',(phi_func,q_func))
        reciprocal_grid.array=np.swapaxes(reciprocal_grid.array,0,1)[...,::-1]
        return reciprocal_grid

    def realGrid():
        def r_func(*args):
            #return (zeros*realCutOff)/zeros[:,-1,None]
            return (zeros)
        real_grid=GridFactory.construct_grid('uniform',(phi_func,r_func))
        real_grid.array=np.swapaxes(real_grid.array,0,1)[...,::-1]
        return real_grid

    ftGridPair=FTGridPair(realGrid(),reciprocalGrid())
    return ftGridPair

def polar_uniform(realCutOff,N1,N2):
    max_order=int((N2-N2%2)/2)
    orders=np.concatenate((np.arange(max_order+1),-1*np.arange(max_order+N2%2)[:0:-1]))
    radial_steps=np.swapaxes(np.repeat(np.arange(N1)[:,None],N2,axis=1),0,1)
    radial_steps=radial_steps+np.abs(orders[:,None])*np.pi/2*0.95 # q min for 50
    def phi_func():
        return orders*np.pi/max_order
    
    def reciprocalGrid():
        def q_func(*args):
            log.info(args[0].shape)
            return (radial_steps)/realCutOff
            
        reciprocal_grid=GridFactory.construct_grid('uniform_dependent',(phi_func,q_func))
        reciprocal_grid.array=np.swapaxes(reciprocal_grid.array,0,1)[...,::-1]
        return reciprocal_grid

    def realGrid():
        def r_func(*args):
            #return (zeros*realCutOff)/zeros[:,-1,None]
            log.info(args[0].shape)
            return (radial_steps*realCutOff/N1)
        real_grid=GridFactory.construct_grid('uniform_dependent',(phi_func,r_func))
        real_grid.array=np.swapaxes(real_grid.array,0,1)[...,::-1]
        return real_grid

    ftGridPair=FTGridPair(realGrid(),reciprocalGrid())
    return ftGridPair
    
def polarFFTGridPair_Bessel_dependend(realCutOff,N1,N2=0):
    def reciprocalGrid():
        def gridPart(zeroNumber):
   #         maxBesselOrder= 2**(np.floor(np.log(zeroNumber)/np.log(2))+1)
            maxBesselOrder= np.floor(np.sqrt(zeroNumber))**2
  #          log.info('maxBesselOrder={}'.format(maxBesselOrder))
            besselOrders=np.arange(np.floor((maxBesselOrder-1)/2),-np.floor((maxBesselOrder-1)/2)-2,-1)
#            log.info('maxBesselOrder={}'.format(maxBesselOrder))
 #           log.info('besselOrders={}'.format(besselOrders))

            repeatedZeroNumber=np.repeat(zeroNumber,maxBesselOrder)
            zeros=mLib.besselZeros(besselOrders,zeroNumber)
            polarSteps=2*np.pi*besselOrders/maxBesselOrder
#            log.info('polarSteps= \n {}'.format(polarSteps))
            gridPart=np.stack((zeros,polarSteps),axis=1)
            return gridPart,int(maxBesselOrder)
        numbersOfZeros=np.arange(1,N1,1)
        gridPartAndShapes=np.array(list(map(gridPart,numbersOfZeros)))
        gridArray=np.concatenate((gridPartAndShapes[:,0]))
        gridShape=[np.array([N1]),gridPartAndShapes[:,1]]
        reziprocalGrid=Grid(gridArray,gridShape)
        return reziprocalGrid

    def realGrid():
        def gridPart(zeroNumber):
#            maxBesselOrder= 2**(np.floor(np.log(zeroNumber)/np.log(2))+2)
            steps=np.array([2,3,5,7])
            maxOrderKandidates=np.floor(np.log(zeroNumber/steps)/np.log(2))
            log.info('Order Kandidates={}'.format(maxOrderKandidates))
            log.info(' selection = {}'.format(np.log(zeroNumber/steps)/np.log(2)-np.floor(np.log(zeroNumber/steps)/np.log(2))))
            minimalindex=np.abs(np.log(zeroNumber/steps)/np.log(2)-np.floor(np.log(zeroNumber/steps)/np.log(2))).argmin()
            maxBesselOrder=steps[minimalindex]*2**maxOrderKandidates[minimalindex]
#            maxBesselOrder= np.floor(np.sqrt(zeroNumber))**2
  #          log.info('maxBesselOrder={}'.format(maxBesselOrder))
            besselOrders=np.arange(np.floor((maxBesselOrder-1)/2),-np.floor((maxBesselOrder-1)/2)-2,-1)
#            log.info('maxBesselOrder={}'.format(maxBesselOrder))
 #           log.info('besselOrders={}'.format(besselOrders))

            repeatedZeroNumber=np.repeat(zeroNumber,maxBesselOrder)
            zeros=mLib.besselZeros(besselOrders,zeroNumber)/mLib.besselZeros(besselOrders,N1)
            polarSteps=2*np.pi*besselOrders/maxBesselOrder
#            log.info('polarSteps= \n {}'.format(polarSteps))
            gridPart=np.stack((zeros,polarSteps),axis=1)
            return gridPart,int(maxBesselOrder)
        numbersOfZeros=np.arange(1,N1,1)
        gridPartAndShapes=np.array(list(map(gridPart,numbersOfZeros)))
        gridArray=np.concatenate((gridPartAndShapes[:,0]))
        gridShape=[np.array([N1]),gridPartAndShapes[:,1]]
        reziprocalGrid=Grid(gridArray,gridShape)
        return reziprocalGrid
    #factory=gridFactory()
    #reziprocalBesselGrid=factory.constructGrid((radialReciprocal,polarReziprocal),factory.type_dependendUniform)
    fftGridPair=FFTGridPair(realGrid(),reciprocalGrid())
    return fftGridPair


def polarFTGridPair_PNAS(reciprocalCutOff,nRadialSteps,nAngularSteps):
    Nr=nRadialSteps
    Nphi=nAngularSteps
    angular_function=uniformGrid_func([0,2*np.pi],nPoints=Nphi,noEndpoint=True)
    real_radial_function=uniformGrid_func([0,1],nPoints=Nr,noEndpoint=False) 
    real_grid=GridFactory.construct_grid('uniform',[real_radial_function,angular_function])

    reciprocal_radial_function=uniformGrid_func([0,reciprocalCutOff],nPoints=Nr,noEndpoint=False) 
    reciprocal_grid=GridFactory.construct_grid('uniform',[reciprocal_radial_function,angular_function])

    FTGrid_pair=FTGridPair(real_grid,reciprocal_grid)
    return FTGrid_pair


def radial_grid_func_zernike_old(reciprocal_cut_off,n_radial_points):
    Nr=n_radial_points
    q_max=reciprocal_cut_off
    r_max= Nr/q_max
    r_step=r_max/Nr
    real_radial_function=uniformGrid_func([1/q_max,r_max],nPoints=Nr,noEndpoint=False)
    reciprocal_radial_function=uniformGrid_func([1/r_max,q_max],nPoints=Nr,noEndpoint=False) 
    return {'real':real_radial_function,'reciprocal':reciprocal_radial_function}

    

def n_radial_points_from_oversampling(max_q,reciprocity_coefficient):
    opt = settings.analysis
    max_r=opt.grid.oversampling*opt.particle_radius
    n_radial_points = int(max_r*max_q/reciprocity_coefficient)+1
    return n_radial_points

def radial_grid_func_zernike(reciprocal_cut_off,n_radial_points,reciprocity_coefficient):
    Nr=n_radial_points
    q_max=reciprocal_cut_off
    r_max = polar_spherical_dft_reciprocity_relation_radial_cutoffs(q_max,Nr,reciprocity_coefficient=reciprocity_coefficient)
    r_step=r_max/Nr
    real_radial_function=uniformGrid_func([0,r_max],nPoints=Nr,noEndpoint=False)
    reciprocal_radial_function=uniformGrid_func([0,q_max],nPoints=Nr,noEndpoint=False) 
    return {'real':real_radial_function,'reciprocal':reciprocal_radial_function}
def radial_grid_func_midpoint(reciprocal_cut_off,n_radial_points,reciprocity_coefficient):
    Nr=n_radial_points
    q_max=reciprocal_cut_off
    r_max = polar_spherical_dft_reciprocity_relation_radial_cutoffs(q_max,Nr,reciprocity_coefficient=reciprocity_coefficient)
    r_step=r_max/Nr
    dr = r_max/Nr
    dq = q_max/Nr
    real_radial_function=np.linspace(0+dr/2,r_max-dr/2,num=Nr,endpoint=True)
    reciprocal_radial_function=np.linspace(0+dq/2,q_max-dq/2,num=Nr,endpoint=True) 
    return {'real':real_radial_function,'reciprocal':reciprocal_radial_function}

def radial_grid_gauss(reciprocal_cut_off,n_radial_points,reciprocity_coefficient):
    Nr=n_radial_points
    q_max=reciprocal_cut_off
    r_max = polar_spherical_dft_reciprocity_relation_radial_cutoffs(q_max,Nr,reciprocity_coefficient=reciprocity_coefficient)
    xs,weights = roots_legendre(Nr)
    rs = r_max/2*xs+r_max/2
    qs = q_max/2*xs+q_max/2
    return {'real':rs,'reciprocal':qs}

def radial_grid_func_PNAS(reciprocal_cut_off,n_radial_points):
    Nr=n_radial_points
    q_max=reciprocal_cut_off
    r_max= 1 # should always sattisfy 1=N/(Q*2)
    r_step=r_max/Nr
    real_radial_function=uniformGrid_func([0,r_max],nPoints=Nr,noEndpoint=False)
    reciprocal_radial_function=uniformGrid_func([0,q_max],nPoints=Nr,noEndpoint=False) 
    return {'real':real_radial_function,'reciprocal':reciprocal_radial_function}
    

def polar_ft_grid_pair_zernike(reciprocal_cut_off,nRadialSteps,nAngularSteps,reciprocity_coefficient):
    radial_func=radial_grid_func_zernike(reciprocal_cut_off,nRadialSteps,reciprocity_coefficient)
    real_radial_function=radial_func['real']
    reciprocal_radial_function=radial_func['reciprocal']

    angular_function=uniformGrid_func([0,2*np.pi],nPoints=nAngularSteps,noEndpoint=True)
#    real_grid=GridFactory.construct_grid('uniform',[np.arange(r_step,r_max,r_step),np.arange(0,2*np.pi,a_step)])
    real_grid=GridFactory.construct_grid('uniform',[real_radial_function,angular_function])
    reciprocal_grid=GridFactory.construct_grid('uniform',[reciprocal_radial_function,angular_function])

    FTGrid_pair=FTGridPair(real_grid,reciprocal_grid)
    return FTGrid_pair

def polar_ft_grid_pair_midpoint(reciprocal_cut_off,nRadialSteps,nAngularSteps,reciprocity_coefficient):
    radial_func=radial_grid_func_midpoint(reciprocal_cut_off,nRadialSteps,reciprocity_coefficient)
    real_radial_function=radial_func['real']
    reciprocal_radial_function=radial_func['reciprocal']

    angular_points=uniformGrid_func([0,2*np.pi],nPoints=nAngularSteps,noEndpoint=True)()
#    real_grid=GridFactory.construct_grid('uniform',[np.arange(r_step,r_max,r_step),np.arange(0,2*np.pi,a_step)])
    real_grid=GridFactory.construct_grid('uniform',[real_radial_function,angular_points])
    reciprocal_grid=GridFactory.construct_grid('uniform',[reciprocal_radial_function,angular_points])

    FTGrid_pair=FTGridPair(real_grid,reciprocal_grid)
    return FTGrid_pair

def polar_ft_grid_pair_gauss(reciprocal_cut_off,nRadialSteps,nAngularSteps,reciprocity_coefficient):
    radial_grids=radial_grid_gauss(reciprocal_cut_off,nRadialSteps,reciprocity_coefficient)
    rs=radial_grids['real']
    qs=radial_grids['reciprocal']

    angular_points=uniformGrid_func([0,2*np.pi],nPoints=nAngularSteps,noEndpoint=True)()
#    real_grid=GridFactory.construct_grid('uniform',[np.arange(r_step,r_max,r_step),np.arange(0,2*np.pi,a_step)])
    real_grid=GridFactory.construct_grid('uniform',[rs,angular_points])
    reciprocal_grid=GridFactory.construct_grid('uniform',[qs,angular_points])

    FTGrid_pair=FTGridPair(real_grid,reciprocal_grid)
    return FTGrid_pair

def spherical_ft_grid_pair_zernike_old(reciprocal_cut_off,nRadialSteps,thetas,phis):
    radial_func=radial_grid_func_zernike_old(reciprocal_cut_off,nRadialSteps)
    rs=radial_func['real']()
    qs=radial_func['reciprocal']()
    if isinstance(thetas,(list,tuple,np.ndarray)) and isinstance(phis,(list,tuple,np.ndarray)): 
        real_grid=GridFactory.construct_grid('uniform',[rs,thetas,phis])
        reciprocal_grid=GridFactory.construct_grid('uniform',[qs,thetas,phis])
    else:
        real_grid=GridFactory.construct_grid('uniform',[rs])
        reciprocal_grid=GridFactory.construct_grid('uniform',[qs])

    FTGrid_pair=FTGridPair(real_grid,reciprocal_grid)
    return FTGrid_pair


def spherical_ft_grid_pair_zernike(reciprocal_cut_off,nRadialSteps,thetas,phis,reciprocity_coefficient):
    radial_func=radial_grid_func_zernike(reciprocal_cut_off,nRadialSteps,reciprocity_coefficient)
    rs=radial_func['real']()
    qs=radial_func['reciprocal']()
    if isinstance(thetas,(list,tuple,np.ndarray)) and isinstance(phis,(list,tuple,np.ndarray)): 
        real_grid=GridFactory.construct_grid('uniform',[rs,thetas,phis])
        reciprocal_grid=GridFactory.construct_grid('uniform',[qs,thetas,phis])
    else:
        real_grid=GridFactory.construct_grid('uniform',[rs])
        reciprocal_grid=GridFactory.construct_grid('uniform',[qs])

    FTGrid_pair=FTGridPair(real_grid,reciprocal_grid)
    return FTGrid_pair

def spherical_ft_grid_pair_midpoint(reciprocal_cut_off,nRadialSteps,thetas,phis,reciprocity_coefficient):
    radial_grid=radial_grid_func_midpoint(reciprocal_cut_off,nRadialSteps,reciprocity_coefficient)
    rs=radial_grid['real']
    qs=radial_grid['reciprocal']
    if isinstance(thetas,(list,tuple,np.ndarray)) and isinstance(phis,(list,tuple,np.ndarray)): 
        real_grid=GridFactory.construct_grid('uniform',[rs,thetas,phis])
        reciprocal_grid=GridFactory.construct_grid('uniform',[qs,thetas,phis])
    else:
        real_grid=GridFactory.construct_grid('uniform',[rs])
        reciprocal_grid=GridFactory.construct_grid('uniform',[qs])

    FTGrid_pair=FTGridPair(real_grid,reciprocal_grid)
    return FTGrid_pair

def spherical_ft_grid_pair_gauss(reciprocal_cut_off,nRadialSteps,thetas,phis,reciprocity_coefficient):
    radial_grid = radial_grid_gauss(reciprocal_cut_off,nRadialSteps,reciprocity_coefficient)
    rs=radial_grid['real']
    qs=radial_grid['reciprocal']
    if isinstance(thetas,(list,tuple,np.ndarray)) and isinstance(phis,(list,tuple,np.ndarray)): 
        real_grid=GridFactory.construct_grid('uniform',[rs,thetas,phis])
        reciprocal_grid=GridFactory.construct_grid('uniform',[qs,thetas,phis])
    else:
        real_grid=GridFactory.construct_grid('uniform',[rs])
        reciprocal_grid=GridFactory.construct_grid('uniform',[qs])

    FTGrid_pair=FTGridPair(real_grid,reciprocal_grid)
    return FTGrid_pair



def spherical_ft_grid_pair_zernike_dict(reciprocal_cut_off,nRadialSteps,thetas,phis,pi_in_q):
    radial_func=radial_grid_func_zernike(reciprocal_cut_off,nRadialSteps,pi_in_q)
    rs=radial_func['real']()
    qs=radial_func['reciprocal']()
    if isinstance(thetas,(list,tuple,np.ndarray)) and isinstance(phis,(list,tuple,np.ndarray)): 
        real_grid=GridFactory.construct_grid('uniform',[rs,thetas,phis])
        reciprocal_grid=GridFactory.construct_grid('uniform',[qs,thetas,phis])
    else:
        real_grid=GridFactory.construct_grid('uniform',[rs])
        reciprocal_grid=GridFactory.construct_grid('uniform',[qs])

    grid_pair = DictNamespace(real=real_grid[:],reciprocal=reciprocal_grid[:])
    return grid_pair

def spherical_ft_grid_pair_SinCos(reciprocal_cut_off,nRadialSteps,thetas,phis):
    radial_func=radial_grid_func_PNAS(reciprocal_cut_off,nRadialSteps)
    rs=radial_func['real']()
    qs=radial_func['reciprocal']()
    if isinstance(thetas,(list,tuple,np.ndarray)) and isinstance(phis,(list,tuple,np.ndarray)): 
        real_grid=GridFactory.construct_grid('uniform',[rs,thetas,phis])
        reciprocal_grid=GridFactory.construct_grid('uniform',[qs,thetas,phis])
    else:
        real_grid=GridFactory.construct_grid('uniform',[rs])
        reciprocal_grid=GridFactory.construct_grid('uniform',[qs])

    FTGrid_pair=FTGridPair(real_grid,reciprocal_grid)
    return FTGrid_pair


def polarFTGridPair_SinCos_new(realCutOff,n_radial_points,n_angular_points):
    reciprocalCutOff=(n_radial_points-1)/realCutOff
    Nr=n_radial_points
    Nphi=n_angular_points
    def realGrid():
        angularFunction=uniformGrid_func([0,2*np.pi],nPoints=Nphi,noEndpoint=True)
        radialFunction=uniformGrid_func([0,realCutOff],nPoints=Nr,noEndpoint=True) 
        realGrid=GridFactory.construct_grid('uniform',[radialFunction,angularFunction])
        return realGrid
    def reciprocalGrid():
        angularFunction=uniformGrid_func([0,2*np.pi],nPoints=Nphi,noEndpoint=True)
        radialFunction=uniformGrid_func([0,reciprocalCutOff],nPoints=Nr,noEndpoint=True) 
        reciprocalGrid=GridFactory.construct_grid('uniform',[radialFunction,angularFunction])
        return reciprocalGrid

    FTGrid_pair=FTGridPair(realGrid(),reciprocalGrid())
    return FTGrid_pair
    

def extractGridParameters_Donatelli(ftGridPair):
    realGrid=ftGridPair.realGrid
    reciprocalGrid=ftGridPair.reciprocalGrid
    gridShape=realGrid.array.shape
    nRadialPoints=gridShape[0]
    nAngularPoints=gridShape[1]

    stepSizes_real=pyLib.uniformGridGetStepSizes(realGrid.array)
    realCutOff=(nRadialPoints-1)*stepSizes_real[0]
    stepSizes_reciprocal=pyLib.uniformGridGetStepSizes(reciprocalGrid.array)
    reciprocalCutOff=(nRadialPoints-1)*stepSizes_reciprocal[0]

    return {'nRadialPoints':nRadialPoints,'nAngularPoints':nAngularPoints,'realCutOff':realCutOff,'reciprocalCutOff':reciprocalCutOff}


#new GridFactory hands full grid part to function woud need to rewrite angularFunction to that.
def polarFFTGridPair_Chebyshev(realCutOff,numberOfRadialNodes):
    def reziprocalGrid():
        radialDomain=[0,numberOfRadialNodes/realCutOff]
        radialFunction=uniformGrid_func(radialDomain,nPoints=numberOfRadialNodes)
        def angularFunction(*args):
            radialPoint=args[0]
            radialPointNumber=args[1]
            if radialPointNumber >=2:
                numberOfPolarNodes=2**(np.floor(np.log(radialPointNumber)/np.log(2))+2)
            elif radialPointNumber==1:
                numberOfPolarNodes=2**2
            else:
                numberOfPolarNodes=1
            polarDomain=np.array([0,2*np.pi])
            func=uniformGrid_func(polarDomain,nPoints=numberOfPolarNodes,noEndpoint=True)
            gridPart=func()
            return gridPart
        grid=GridFactory.construct_grid('dependent',(radialFunction,angularFunction))
        return grid
    def realGrid():
        radialDomain=[0,numberOfRadialNodes/realCutOff]
        radialFunction=uniformGrid_func(radialDomain,nPoints=numberOfRadialNodes)
        def angularFunction(*args):
            radialPoint=args[0]
            radialPointNumber=args[1]
            if radialPointNumber >=2:
                numberOfPolarNodes=2**(np.floor(np.log(radialPointNumber)/np.log(2))+2)
            elif radialPointNumber==1:
                numberOfPolarNodes=2**2
            else:
                numberOfPolarNodes=1
            polarDomain=np.array([0,2*np.pi])
            func=uniformGrid_func(polarDomain,nPoints=numberOfPolarNodes,noEndpoint=True)
            gridPart=func()
            return gridPart
        grid=GridFactory.construct_grid('dependend',(radialFunction,angularFunction))
        return grid
    grid=reziprocalGrid()
    return grid


def get_grid(grid_opt):
    dim=grid_opt['dimensions']
    _type=grid_opt['type']
    #pi_in_q=grid_opt['pi_in_q']
    reciprocity_coefficient = _get_reciprocity_coefficient(grid_opt)
    Q=grid_opt['max_q']
    n_radial_points = grid_opt['n_radial_points']
    #log.info('dim = {},grid_type = {},pi_in_q = {},max_q = {}, n_radial_points = {}'.format(dim,_type,pi_in_q,Q,n_radial_points))
    if isinstance(n_radial_points,int) and not isinstance(n_radial_points,bool):
        N_r = n_radial_points
    elif isinstance(n_radial_points,str):
        N_r = grid_opt['n_radial_points_from_data']
    else:
        N_r = n_radial_points_from_oversampling(Q,reciprocity_coefficient)
        
    if dim == 2:
        #log.info("grid_opt keys = {}".format(grid_opt.keys()))
        N_phi=len(grid_opt['phis'])
        if (_type == 'Zernike') or (_type == 'trapz'):
            grid = polar_ft_grid_pair_zernike(Q,N_r,N_phi,reciprocity_coefficient)
        elif _type == 'PNAS':
            grid = polarFTGridPair_PNAS(Q,N_r,N_phi)
        elif _type == 'midpoint':
            grid = polar_ft_grid_pair_midpoint(Q,N_r,N_phi,reciprocity_coefficient)
        elif _type == 'gauss':
            grid = polar_ft_grid_pair_gauss(Q,N_r,N_phi,reciprocity_coefficient)
    elif dim == 3:
        thetas=grid_opt['thetas']
        phis=grid_opt['phis']
        #log.info('grid type = {}'.format(_type))
        if (_type == 'Zernike') or (_type == 'trapz'):
            grid = spherical_ft_grid_pair_zernike(Q,N_r,thetas,phis,reciprocity_coefficient)
        elif _type == 'PNAS':
            grid = spherical_ft_grid_pair_SinCos(Q,N_r,thetas,phis)
        elif _type == 'midpoint':
            grid = spherical_ft_grid_pair_midpoint(Q,N_r,thetas,phis,reciprocity_coefficient)
        elif _type == 'gauss':
            grid = spherical_ft_grid_pair_gauss(Q,N_r,thetas,phis,reciprocity_coefficient)

    return grid


def get_polar_fft_angles_from_max_order(max_order):
    return np.arange(0,max_order*2)*np.pi/max_order
